import pathlib

from mopidy.internal import deprecation


def path_to_data_dir(name):
    path = pathlib.Path(__file__).parent / "data" / name
    return path.resolve()


def generate_song(i):
    return "local:track:song%s.wav" % i


def populate_tracklist(func):
    def wrapper(self):
        with deprecation.ignore("core.tracklist.add:tracks_arg"):
            self.tl_tracks = self.core.tracklist.add(self.tracks)
        return func(self)

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper


class IsA:
    def __init__(self, klass):
        self.klass = klass

    def __eq__(self, rhs):
        try:
            return isinstance(rhs, self.klass)
        except TypeError:
            return type(rhs) == type(self.klass)  # noqa

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __repr__(self):
        return str(self.klass)



import pathlib

import pkg_resources

from mopidy import config, ext

__version__ = pkg_resources.get_distribution("Mopidy-Local").version


class Extension(ext.Extension):

    dist_name = "Mopidy-Local"
    ext_name = "local"
    version = __version__

    def get_default_config(self):
        return config.read(pathlib.Path(__file__).parent / "ext.conf")

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["library"] = config.Deprecated()
        schema["max_search_results"] = config.Integer(minimum=0)
        schema["media_dir"] = config.Path()
        schema["data_dir"] = config.Deprecated()
        schema["playlists_dir"] = config.Deprecated()
        schema["tag_cache_file"] = config.Deprecated()
        schema["scan_timeout"] = config.Integer(minimum=1000, maximum=1000 * 60 * 60)
        schema["scan_flush_threshold"] = config.Integer(minimum=0)
        schema["scan_follow_symlinks"] = config.Boolean()
        schema["included_file_extensions"] = config.List(optional=True)
        schema["excluded_file_extensions"] = config.List(optional=True)
        schema["directories"] = config.List()
        schema["timeout"] = config.Integer(optional=True, minimum=1)
        schema["use_artist_sortname"] = config.Boolean()
        schema["album_art_files"] = config.List(optional=True)
        return schema

    def setup(self, registry):
        from .actor import LocalBackend

        registry.add("backend", LocalBackend)
        registry.add("http:app", {"name": self.ext_name, "factory": self.webapp})

    def get_command(self):
        from .commands import LocalCommand

        return LocalCommand()

    def webapp(self, config, core):
        from .web import ImageHandler, IndexHandler

        image_dir = self.get_image_dir(config)
        return [
            (r"/(index.html)?", IndexHandler, {"root": image_dir}),
            (r"/(.+)", ImageHandler, {"path": image_dir}),
        ]

    # TODO: Add *paths to Extension.get_data_dir()?
    @classmethod
    def get_data_subdir(cls, config, *paths):
        data_dir = cls.get_data_dir(config)
        dir_path = data_dir.joinpath(*paths)
        dir_path.mkdir(parents=True, exist_ok=True)
        return dir_path

    @classmethod
    def get_image_dir(cls, config):
        return cls.get_data_subdir(config, "images")



import logging
import os
import pathlib

import tornado.web

logger = logging.getLogger(__name__)


class ImageHandler(tornado.web.StaticFileHandler):
    def get_cache_time(self, *args):
        return self.CACHE_MAX_AGE


class IndexHandler(tornado.web.RequestHandler):
    def initialize(self, root):
        self.root = root

    def get(self, path):
        return self.render("index.html", images=self.uris())

    def get_template_path(self):
        return pathlib.Path(__file__).parent / "www"

    def uris(self):
        for _, _, files in os.walk(self.root):
            yield from files



"""A dummy audio actor for use in tests.

This class implements the audio API in the simplest way possible. It is used in
tests of the core and backends.
"""


import pykka

from mopidy import audio


def create_proxy(config=None, mixer=None):
    return DummyAudio.start(config, mixer).proxy()


# TODO: reset position on track change?
class DummyAudio(pykka.ThreadingActor):
    def __init__(self, config=None, mixer=None):
        super().__init__()
        self.state = audio.PlaybackState.STOPPED
        self._volume = 0
        self._position = 0
        self._callback = None
        self._uri = None
        self._stream_changed = False
        self._tags = {}
        self._bad_uris = set()

    def set_uri(self, uri, live_stream=False):
        assert self._uri is None, "prepare change not called before set"
        self._position = 0
        self._uri = uri
        self._stream_changed = True
        self._tags = {}

    def set_appsrc(self, *args, **kwargs):
        pass

    def emit_data(self, buffer_):
        pass

    def get_position(self):
        return self._position

    def set_position(self, position):
        self._position = position
        audio.AudioListener.send("position_changed", position=position)
        return True

    def start_playback(self):
        return self._change_state(audio.PlaybackState.PLAYING)

    def pause_playback(self):
        return self._change_state(audio.PlaybackState.PAUSED)

    def prepare_change(self):
        self._uri = None
        return True

    def stop_playback(self):
        return self._change_state(audio.PlaybackState.STOPPED)

    def get_volume(self):
        return self._volume

    def set_volume(self, volume):
        self._volume = volume
        return True

    def set_metadata(self, track):
        pass

    def get_current_tags(self):
        return self._tags

    def set_about_to_finish_callback(self, callback):
        self._callback = callback

    def enable_sync_handler(self):
        pass

    def wait_for_state_change(self):
        pass

    def _change_state(self, new_state):
        if not self._uri:
            return False

        if new_state == audio.PlaybackState.STOPPED and self._uri:
            self._stream_changed = True
            self._uri = None

        if self._uri is not None:
            audio.AudioListener.send("position_changed", position=0)

        if self._stream_changed:
            self._stream_changed = False
            audio.AudioListener.send("stream_changed", uri=self._uri)

        old_state, self.state = self.state, new_state
        audio.AudioListener.send(
            "state_changed", old_state=old_state, new_state=new_state, target_state=None
        )

        if new_state == audio.PlaybackState.PLAYING:
            self._tags["audio-codec"] = ["fake info..."]
            audio.AudioListener.send("tags_changed", tags=["audio-codec"])

        return self._uri not in self._bad_uris

    def trigger_fake_playback_failure(self, uri):
        self._bad_uris.add(uri)

    def trigger_fake_tags_changed(self, tags):
        self._tags.update(tags)
        audio.AudioListener.send("tags_changed", tags=self._tags.keys())

    def get_about_to_finish_callback(self):
        # This needs to be called from outside the actor or we lock up.
        def wrapper():
            if self._callback:
                self.prepare_change()
                self._callback()

            if not self._uri or not self._callback:
                self._tags = {}
                audio.AudioListener.send("reached_end_of_stream")
            else:
                audio.AudioListener.send("position_changed", position=0)
                audio.AudioListener.send("stream_changed", uri=self._uri)

        return wrapper



import hashlib
import imghdr
import logging
import pathlib
import shutil
import sqlite3
import struct

import uritools

from . import Extension, schema, translator

logger = logging.getLogger(__name__)


def check_dirs_and_files(config):
    if not pathlib.Path(config["local"]["media_dir"]).is_dir():
        logger.warning(
            "Local media dir %s does not exist or we lack permissions to the "
            "directory or one of its parents" % config["local"]["media_dir"]
        )


# would be nice to have these in imghdr...
def get_image_size_png(data):
    return struct.unpack(">ii", data[16:24])


def get_image_size_gif(data):
    return struct.unpack("<HH", data[6:10])


def model_uri(type, model):
    if type == "album":
        # ignore num_tracks for multi-disc albums
        digest = hashlib.md5(str(model.replace(num_tracks=None)).encode())
    else:
        digest = hashlib.md5(str(model).encode())
    return "local:{}:md5:{}".format(type, digest.hexdigest())


def get_image_size_jpeg(data):
    # original source: http://goo.gl/6bo5Vx
    index = 0
    ftype = 0
    size = 2
    while not 0xC0 <= ftype <= 0xCF:
        index += size
        ftype = data[index]
        while ftype == 0xFF:
            index += 1
            ftype = data[index]
        index += 1
        size = struct.unpack(">H", data[index : index + 2])[0] - 2
        index += 2
    index += 1  # skip precision byte
    height, width = struct.unpack(">HH", data[index : index + 4])
    return width, height


class LocalStorageProvider:
    def __init__(self, config):
        self._config = ext_config = config[Extension.ext_name]
        self._media_dir = pathlib.Path(ext_config["media_dir"])
        self._data_dir = Extension.get_data_dir(config)
        self._image_dir = Extension.get_image_dir(config)
        self._base_uri = "/" + Extension.ext_name + "/"
        self._patterns = list(map(str, ext_config["album_art_files"]))
        self._dbpath = self._data_dir / "library.db"
        self._connection = None

    def load(self):
        with self._connect() as connection:
            version = schema.load(connection)
            logger.debug("Using SQLite database schema v%s", version)
            return schema.count_tracks(connection)

    def begin(self):
        return schema.tracks(self._connect())

    def add(self, track, tags=None, duration=None):
        logger.debug("Adding track: %s", track)
        images = None
        if track.album and track.album.name:  # FIXME: album required
            uri = translator.local_uri_to_file_uri(track.uri, self._media_dir)
            try:
                images = self._extract_images(track.uri, tags)
                logger.debug("%s images: %s", track.uri, images)
            except Exception as e:
                logger.warning("Error extracting images for %s: %s", uri, e)
        try:
            track = self._validate_track(track)
            schema.insert_track(self._connect(), track, images)
        except Exception as e:
            logger.warning("Skipped %s: %s", track.uri, e)

    def remove(self, uri):
        schema.delete_track(self._connect(), uri)

    def flush(self):
        if not self._connection:
            return False
        self._connection.commit()
        return True

    def close(self):
        if self._connection:
            schema.cleanup(self._connection)
            self._connection.commit()
            self._connection.close()
            self._connection = None
        else:
            logger.error("Attempting to close while not connected")
        self._cleanup_images()

    def clear(self):
        logger.info("Clearing image directory")
        try:
            shutil.rmtree(self._image_dir)
            self._image_dir.mkdir()
        except IOError as e:
            logger.warning("Error clearing image directory: %s", e)
        logger.info("Clearing SQLite database")
        try:
            schema.clear(self._connect())
            return True
        except sqlite3.Error as e:
            logger.error("Error clearing SQLite database: %s", e)
            return False

    def _connect(self):
        if not self._connection:
            self._connection = sqlite3.connect(
                self._dbpath,
                factory=schema.Connection,
                timeout=self._config["timeout"],
                check_same_thread=False,
            )
        return self._connection

    def _validate_artist(self, model):
        if not model.name:
            raise ValueError("Empty artist name")
        if not model.uri:
            model = model.replace(uri=model_uri("artist", model))
        return model

    def _validate_album(self, model):
        if not model.name:
            raise ValueError("Empty album name")
        if not model.uri:
            model = model.replace(uri=model_uri("album", model))
        return model.replace(artists=list(map(self._validate_artist, model.artists)))

    def _validate_track(self, model):
        if not model.uri:
            raise ValueError("Empty track URI")
        if model.name:
            name = model.name
        else:
            name = translator.local_uri_to_path(model.uri, "").name
        if model.album and model.album.name:
            album = self._validate_album(model.album)
        else:
            album = None
        return model.replace(
            name=name,
            album=album,
            artists=list(map(self._validate_artist, model.artists)),
            composers=list(map(self._validate_artist, model.composers)),
            performers=list(map(self._validate_artist, model.performers)),
        )

    def _cleanup_images(self):
        logger.info("Cleaning up image directory")
        with self._connect() as c:
            uris = set(schema.get_image_uris(c))
        for image_path in self._image_dir.glob("**/*"):
            if uritools.urijoin(self._base_uri, image_path.name) not in uris:
                logger.info(f"Deleting file {image_path.as_uri()}")
                image_path.unlink()

    def _extract_images(self, uri, tags):
        images = set()  # filter duplicate images, e.g. embedded/external
        for image in tags.get("image", []) + tags.get("preview-image", []):
            try:
                # FIXME: gst.Buffer or plain str/bytes type?
                data = getattr(image, "data", image)
                images.add(self._get_or_create_image_file(None, data))
            except Exception as e:
                logger.warning("Error extracting images for %r: %r", uri, e)
        # look for external album art
        track_path = translator.local_uri_to_path(uri, self._media_dir)
        dir_path = track_path.parent
        for pattern in self._patterns:
            for match_path in dir_path.glob(pattern):
                try:
                    images.add(self._get_or_create_image_file(match_path))
                except Exception as e:
                    logger.warning(
                        f"Cannot read image file {match_path.as_uri()}: {e!r}"
                    )
        return images

    def _get_or_create_image_file(self, path, data=None):
        what = imghdr.what(path, data)
        if not what:
            raise ValueError("Unknown image type")
        if not data:
            with open(path, "rb") as f:
                data = f.read()
        digest, width, height = hashlib.md5(data).hexdigest(), None, None
        try:
            if what == "png":
                width, height = get_image_size_png(data)
            elif what == "gif":
                width, height = get_image_size_gif(data)
            elif what == "jpeg":
                width, height = get_image_size_jpeg(data)
        except Exception as e:
            logger.error("Error getting image size for %r: %r", path, e)
        if width and height:
            name = "%s-%dx%d.%s" % (digest, width, height, what)
        else:
            name = f"{digest}.{what}"
        image_path = self._image_dir / name
        if not image_path.is_file():
            logger.info(f"Creating file {image_path.as_uri()}")
            image_path.write_bytes(data)
        return uritools.urijoin(self._base_uri, name)



import os
import pathlib
import shutil

import pytest

from mopidy import exceptions
from mopidy_local import mtimes

import tests


@pytest.fixture
def tmp_dir_path(tmp_path):
    yield tmp_path
    if tmp_path.is_dir():
        shutil.rmtree(str(tmp_path))


def test_find_error_is_a_mopidy_exception():
    assert issubclass(mtimes.FindError, exceptions.MopidyException)


def test_find_error_can_store_an_errno():
    exc = mtimes.FindError("msg", errno=1234)

    assert exc.message == "msg"
    assert exc.errno == 1234


def test_names_are_pathlib_objects():
    result, errors = mtimes.find_mtimes(str(tests.path_to_data_dir("")))

    for name in list(result.keys()) + list(errors.keys()):
        assert isinstance(name, pathlib.Path)


def test_nonexistent_dir_is_an_error(tmp_dir_path):
    missing_path = tmp_dir_path / "does-not-exist"

    result, errors = mtimes.find_mtimes(missing_path)

    assert result == {}
    assert errors == {missing_path: tests.IsA(mtimes.FindError)}


def test_empty_dirs_are_not_in_the_result(tmp_dir_path):
    """Empty directories should not show up in results"""
    dir_path = tmp_dir_path / "empty"
    dir_path.mkdir()

    result, errors = mtimes.find_mtimes(dir_path)

    assert result == {}
    assert errors == {}


def test_file_as_the_root_just_returns_the_file(tmp_dir_path):
    file_path = tmp_dir_path / "single"
    file_path.touch()

    result, errors = mtimes.find_mtimes(file_path)

    assert result == {file_path: tests.IsA(int)}
    assert errors == {}


def test_nested_directories(tmp_dir_path):
    # Setup foo/bar and baz directories
    foo_path = tmp_dir_path / "foo" / "file"
    foo_path.parent.mkdir()
    foo_path.touch()
    foo_bar_path = tmp_dir_path / "foo" / "bar" / "filee"
    foo_bar_path.parent.mkdir()
    foo_bar_path.touch()
    baz_path = tmp_dir_path / "baz" / "file"
    baz_path.parent.mkdir()
    baz_path.touch()

    result, errors = mtimes.find_mtimes(tmp_dir_path)

    assert result == {
        foo_path: tests.IsA(int),
        foo_bar_path: tests.IsA(int),
        baz_path: tests.IsA(int),
    }
    assert errors == {}


def test_missing_permission_to_file_is_not_an_error(tmp_dir_path):
    """Missing permissions to a file is not a search error"""
    file_path = tmp_dir_path / "file"
    file_path.touch(mode=0o000)

    result, errors = mtimes.find_mtimes(tmp_dir_path)

    assert result == {file_path: tests.IsA(int)}
    assert errors == {}

    file_path.chmod(0o644)


def test_missing_permission_to_directory_is_an_error(tmp_dir_path):
    dir_path = tmp_dir_path / "dir"
    dir_path.mkdir(mode=0o000)

    result, errors = mtimes.find_mtimes(tmp_dir_path)

    assert result == {}
    assert errors == {dir_path: tests.IsA(mtimes.FindError)}

    dir_path.chmod(0o755)


def test_symlinks_are_by_default_an_error(tmp_dir_path):
    """By default symlinks should be treated as an error"""
    file_path = tmp_dir_path / "file"
    file_path.touch()
    link_path = tmp_dir_path / "link"
    link_path.symlink_to(file_path)

    result, errors = mtimes.find_mtimes(tmp_dir_path)

    assert result == {file_path: tests.IsA(int)}
    assert errors == {link_path: tests.IsA(mtimes.FindError)}


def test_with_follow_symlink_to_file_as_root_is_followed(tmp_dir_path):
    file_path = tmp_dir_path / "file"
    file_path.touch()
    link_path = tmp_dir_path / "link"
    link_path.symlink_to(file_path)

    result, errors = mtimes.find_mtimes(link_path, follow=True)

    assert result == {file_path: tests.IsA(int)}
    assert errors == {}


def test_symlink_to_directory_is_followed(tmp_dir_path):
    file_path = tmp_dir_path / "dir" / "file"
    file_path.parent.mkdir()
    file_path.touch()
    link_path = tmp_dir_path / "link"
    link_path.symlink_to(file_path.parent, target_is_directory=True)

    result, errors = mtimes.find_mtimes(link_path, follow=True)

    assert result == {file_path: tests.IsA(int)}
    assert errors == {}


def test_symlink_pointing_at_itself_fails(tmp_dir_path):
    link_path = tmp_dir_path / "link"
    link_path.symlink_to(link_path)

    result, errors = mtimes.find_mtimes(tmp_dir_path, follow=True)

    assert result == {}
    assert errors == {link_path: tests.IsA(mtimes.FindError)}


def test_symlink_pointing_at_parent_fails(tmp_dir_path):
    """We should detect a loop via the parent and give up on the branch"""

    link_path = tmp_dir_path / "link"
    link_path.symlink_to(tmp_dir_path, target_is_directory=True)

    result, errors = mtimes.find_mtimes(tmp_dir_path, follow=True)

    assert result == {}
    assert errors == {link_path: tests.IsA(Exception)}


def test_indirect_symlink_loop(tmp_dir_path):
    """More indirect loops should also be detected"""
    # Setup tmpdir/directory/loop where loop points to tmpdir
    link_path = tmp_dir_path / "dir" / "link"
    link_path.parent.mkdir()
    link_path.symlink_to(tmp_dir_path, target_is_directory=True)

    result, errors = mtimes.find_mtimes(tmp_dir_path, follow=True)

    assert result == {}
    assert errors == {link_path: tests.IsA(Exception)}


def test_symlink_branches_are_not_excluded(tmp_dir_path):
    """Using symlinks to make a file show up multiple times should work"""
    file_path = tmp_dir_path / "dir" / "file"
    file_path.parent.mkdir()
    file_path.touch()
    link1_path = tmp_dir_path / "link1"
    link1_path.symlink_to(file_path)
    link2_path = tmp_dir_path / "link2"
    link2_path.symlink_to(file_path)

    result, errors = mtimes.find_mtimes(tmp_dir_path, follow=True)

    assert result == {
        file_path: tests.IsA(int),
        link1_path: tests.IsA(int),
        link2_path: tests.IsA(int),
    }
    assert errors == {}


def test_gives_mtime_in_milliseconds(tmp_dir_path):
    file_path = tmp_dir_path / "file"
    file_path.touch()

    os.utime(str(file_path), (1, 3.14159265))

    result, errors = mtimes.find_mtimes(tmp_dir_path)

    assert result == {file_path: 3141}
    assert errors == {}



import logging
import pathlib
import time

from mopidy import commands, exceptions
from mopidy.audio import scan, tags

from mopidy_local import mtimes, storage, translator

logger = logging.getLogger(__name__)

MIN_DURATION_MS = 100  # Shortest length of track to include.


class LocalCommand(commands.Command):
    def __init__(self):
        super().__init__()
        self.add_child("scan", ScanCommand())
        self.add_child("clear", ClearCommand())


class ClearCommand(commands.Command):
    help = "Clear local media files from the local library."

    def run(self, args, config):
        library = storage.LocalStorageProvider(config)

        prompt = "Are you sure you want to clear the library? [y/N] "

        if input(prompt).lower() != "y":
            print("Clearing library aborted")
            return 0

        if library.clear():
            print("Library successfully cleared")
            return 0

        print("Unable to clear library")
        return 1


class ScanCommand(commands.Command):
    help = "Scan local media files and populate the local library."

    def __init__(self):
        super().__init__()
        self.add_argument(
            "--limit",
            action="store",
            type=int,
            dest="limit",
            default=None,
            help="Maximum number of tracks to scan",
        )
        self.add_argument(
            "--force",
            action="store_true",
            dest="force",
            default=False,
            help="Force rescan of all media files",
        )

    def run(self, args, config):
        media_dir = pathlib.Path(config["local"]["media_dir"]).resolve()
        library = storage.LocalStorageProvider(config)

        file_mtimes = self._find_files(
            media_dir=media_dir,
            follow_symlinks=config["local"]["scan_follow_symlinks"],
        )

        files_to_update, files_in_library = self._check_tracks_in_library(
            media_dir=media_dir,
            file_mtimes=file_mtimes,
            library=library,
            force_rescan=args.force,
        )

        files_to_update.update(
            self._find_files_to_scan(
                media_dir=media_dir,
                file_mtimes=file_mtimes,
                files_in_library=files_in_library,
                included_file_exts=[
                    file_ext.lower()
                    for file_ext in config["local"]["included_file_extensions"]
                ],
                excluded_file_exts=[
                    file_ext.lower()
                    for file_ext in config["local"]["excluded_file_extensions"]
                ],
            )
        )

        self._scan_metadata(
            media_dir=media_dir,
            file_mtimes=file_mtimes,
            files=files_to_update,
            library=library,
            timeout=config["local"]["scan_timeout"],
            flush_threshold=config["local"]["scan_flush_threshold"],
            limit=args.limit,
        )

        library.close()
        return 0

    def _find_files(self, *, media_dir, follow_symlinks):
        logger.info(f"Finding files in {media_dir.as_uri()} ...")
        file_mtimes, file_errors = mtimes.find_mtimes(media_dir, follow=follow_symlinks)
        logger.info(f"Found {len(file_mtimes)} files in {media_dir.as_uri()}")

        if file_errors:
            logger.warning(
                f"Encountered {len(file_errors)} errors "
                f"while finding files in {media_dir.as_uri()}"
            )
        for path in file_errors:
            logger.warning(f"Error for {path.as_uri()}: {file_errors[path]}")

        return file_mtimes

    def _check_tracks_in_library(
        self, *, media_dir, file_mtimes, library, force_rescan
    ):
        num_tracks = library.load()
        logger.info(f"Checking {num_tracks} tracks from library")

        uris_to_remove = set()
        files_to_update = set()
        files_in_library = set()

        for track in library.begin():
            absolute_path = translator.local_uri_to_path(track.uri, media_dir)
            mtime = file_mtimes.get(absolute_path)
            if mtime is None:
                logger.debug(f"Removing {track.uri}: File not found")
                uris_to_remove.add(track.uri)
            elif mtime > track.last_modified or force_rescan:
                files_to_update.add(absolute_path)
            files_in_library.add(absolute_path)

        logger.info(f"Removing {len(uris_to_remove)} missing tracks")
        for local_uri in uris_to_remove:
            library.remove(local_uri)

        return files_to_update, files_in_library

    def _find_files_to_scan(
        self,
        *,
        media_dir,
        file_mtimes,
        files_in_library,
        included_file_exts,
        excluded_file_exts,
    ):
        files_to_update = set()

        def _is_hidden_file(relative_path, file_uri):
            if any(p.startswith(".") for p in relative_path.parts):
                logger.debug(f"Skipped {file_uri}: Hidden directory/file")
                return True
            else:
                return False

        def _extension_filters(
            relative_path, file_uri, included_file_exts, excluded_file_exts
        ):
            if included_file_exts:
                if relative_path.suffix.lower() in included_file_exts:
                    logger.debug(f"Added {file_uri}: File extension on included list")
                    return True
                else:
                    logger.debug(
                        f"Skipped {file_uri}: File extension not on included list"
                    )
                    return False
            else:
                if relative_path.suffix.lower() in excluded_file_exts:
                    logger.debug(f"Skipped {file_uri}: File extension on excluded list")
                    return False
                else:
                    logger.debug(
                        f"Included {file_uri}: File extension not on excluded list"
                    )
                    return True

        for absolute_path in file_mtimes:
            relative_path = absolute_path.relative_to(media_dir)
            file_uri = absolute_path.as_uri()

            if (
                not _is_hidden_file(relative_path, file_uri)
                and _extension_filters(
                    relative_path, file_uri, included_file_exts, excluded_file_exts
                )
                and absolute_path not in files_in_library
            ):
                files_to_update.add(absolute_path)

        logger.info(f"Found {len(files_to_update)} tracks which need to be updated")
        return files_to_update

    def _scan_metadata(
        self, *, media_dir, file_mtimes, files, library, timeout, flush_threshold, limit
    ):
        logger.info("Scanning...")

        files = sorted(files)[:limit]

        scanner = scan.Scanner(timeout)
        progress = _ScanProgress(batch_size=flush_threshold, total=len(files))

        for absolute_path in files:
            try:
                file_uri = absolute_path.as_uri()
                result = scanner.scan(file_uri)

                if not result.playable:
                    logger.warning(
                        f"Failed scanning {file_uri}: No audio found in file"
                    )
                elif result.duration is None:
                    logger.warning(
                        f"Failed scanning {file_uri}: "
                        "No duration information found in file"
                    )
                elif result.duration < MIN_DURATION_MS:
                    logger.warning(
                        f"Failed scanning {file_uri}: "
                        f"Track shorter than {MIN_DURATION_MS}ms"
                    )
                else:
                    local_uri = translator.path_to_local_track_uri(
                        absolute_path, media_dir
                    )
                    mtime = file_mtimes.get(absolute_path)
                    track = tags.convert_tags_to_track(result.tags).replace(
                        uri=local_uri, length=result.duration, last_modified=mtime
                    )
                    library.add(track, result.tags, result.duration)
                    logger.debug(f"Added {track.uri}")
            except exceptions.ScannerError as error:
                logger.warning(f"Failed scanning {file_uri}: {error}")

            if progress.increment():
                progress.log()
                if library.flush():
                    logger.debug("Progress flushed")

        progress.log()
        logger.info("Done scanning")


class _ScanProgress:
    def __init__(self, *, batch_size, total):
        self.count = 0
        self.batch_size = batch_size
        self.total = total
        self.start = time.time()

    def increment(self):
        self.count += 1
        return self.batch_size and self.count % self.batch_size == 0

    def log(self):
        duration = time.time() - self.start
        if self.count >= self.total or not self.count:
            logger.info(
                f"Scanned {self.count} of {self.total} files in {duration:.3f}s."
            )
        else:
            remainder = duration / self.count * (self.total - self.count)
            logger.info(
                f"Scanned {self.count} of {self.total} files "
                f"in {duration:.3f}s, ~{remainder:.0f}s left"
            )



import logging

import pykka

from mopidy import backend

from mopidy_local import storage
from mopidy_local.library import LocalLibraryProvider
from mopidy_local.playback import LocalPlaybackProvider

logger = logging.getLogger(__name__)


class LocalBackend(pykka.ThreadingActor, backend.Backend):
    uri_schemes = ["local"]

    def __init__(self, config, audio):
        super().__init__()

        self.config = config

        storage.check_dirs_and_files(config)

        self.playback = LocalPlaybackProvider(audio=audio, backend=self)
        self.library = LocalLibraryProvider(backend=self, config=config)



import pathlib
import queue
import stat
import threading

from mopidy import exceptions


class FindError(exceptions.MopidyException):
    def __init__(self, message, errno=None):
        super().__init__(message, errno)
        self.errno = errno


def find_mtimes(root, follow=False):
    results, errors = _find(root, relative=False, follow=follow)

    # return the mtimes as integer milliseconds
    mtimes = {f: int(st.st_mtime * 1000) for f, st in results.items()}

    return mtimes, errors


def _find(root, thread_count=10, relative=False, follow=False):
    """Threaded find implementation that provides stat results for files.

    Tries to protect against sym/hardlink loops by keeping an eye on parent
    (st_dev, st_ino) pairs.

    :param Path root: root directory to search from, may not be a file
    :param int thread_count: number of workers to use, mainly useful to
        mitigate network lag when scanning on NFS etc.
    :param bool relative: if results should be relative to root or absolute
    :param bool follow: if symlinks should be followed
    """
    root = pathlib.Path(root).resolve()
    threads = []
    results = {}
    errors = {}
    done = threading.Event()
    work = queue.Queue()
    work.put((root, []))

    if not relative:
        root = None

    args = (root, follow, done, work, results, errors)
    for _ in range(thread_count):
        t = threading.Thread(target=_find_worker, args=args)
        t.daemon = True
        t.start()
        threads.append(t)

    work.join()
    done.set()
    for t in threads:
        t.join()
    return results, errors


def _find_worker(root, follow, done, work, results, errors):
    """Worker thread for collecting stat() results.

    :param Path root: directory to make results relative to
    :param bool follow: if symlinks should be followed
    :param threading.Event done: event indicating that all work has been done
    :param queue.Queue work: queue of paths to process
    :param dict results: shared dictionary for storing all the stat() results
    :param dict errors: shared dictionary for storing any per path errors
    """
    while not done.is_set():
        try:
            entry, parents = work.get(block=False)
        except queue.Empty:
            continue

        if root:
            path = entry.relative_to(root)
        else:
            path = entry

        try:
            if follow:
                st = entry.stat()
            else:
                st = entry.lstat()

            if (st.st_dev, st.st_ino) in parents:
                errors[path] = FindError("Sym/hardlink loop found.")
                continue

            if stat.S_ISDIR(st.st_mode):
                for e in entry.iterdir():
                    work.put((e, parents + [(st.st_dev, st.st_ino)]))
            elif stat.S_ISREG(st.st_mode):
                results[path] = st
            elif stat.S_ISLNK(st.st_mode):
                errors[path] = FindError("Not following symlinks.")
            else:
                errors[path] = FindError("Not a file or directory.")

        except OSError as exc:
            errors[path] = FindError(exc.strerror, exc.errno)
        finally:
            work.task_done()



import logging
import operator
import sqlite3

import uritools
from mopidy import backend, models
from mopidy.models import Ref, SearchResult

from . import Extension, schema

logger = logging.getLogger(__name__)


def date_ref(date):
    return Ref.directory(
        uri=uritools.uricompose("local", None, "directory", {"date": date}), name=date
    )


def genre_ref(genre):
    return Ref.directory(
        uri=uritools.uricompose("local", None, "directory", {"genre": genre}),
        name=genre,
    )


class LocalLibraryProvider(backend.LibraryProvider):

    ROOT_DIRECTORY_URI = "local:directory"

    root_directory = models.Ref.directory(uri=ROOT_DIRECTORY_URI, name="Local media")

    def __init__(self, backend, config):
        super().__init__(backend)
        self._config = ext_config = config[Extension.ext_name]
        self._data_dir = Extension.get_data_dir(config)
        self._directories = []
        for line in ext_config["directories"]:
            name, uri = line.rsplit(None, 1)
            ref = Ref.directory(uri=uri, name=name)
            self._directories.append(ref)
        self._dbpath = self._data_dir / "library.db"
        self._connection = None

    def load(self):
        with self._connect() as connection:
            version = schema.load(connection)
            logger.debug("Using SQLite database schema v%s", version)
            return schema.count_tracks(connection)

    def lookup(self, uri):
        try:
            if uri.startswith("local:album"):
                return list(schema.lookup(self._connect(), Ref.ALBUM, uri))
            elif uri.startswith("local:artist"):
                return list(schema.lookup(self._connect(), Ref.ARTIST, uri))
            elif uri.startswith("local:track"):
                return list(schema.lookup(self._connect(), Ref.TRACK, uri))
            else:
                raise ValueError("Invalid lookup URI")
        except Exception as e:
            logger.error("Lookup error for %s: %s", uri, e)
            return []

    def browse(self, uri):
        try:
            if uri == self.ROOT_DIRECTORY_URI:
                return self._directories
            elif uri.startswith("local:directory"):
                return self._browse_directory(uri)
            elif uri.startswith("local:artist"):
                return self._browse_artist(uri)
            elif uri.startswith("local:album"):
                return self._browse_album(uri)
            else:
                raise ValueError("Invalid browse URI")
        except Exception as e:
            logger.error("Error browsing %s: %s", uri, e)
            return []

    def search(self, query=None, limit=100, offset=0, uris=None, exact=False):
        limit = self._config["max_search_results"]
        q = []
        for field, values in query.items() if query else []:
            q.extend((field, value) for value in values)
        filters = [f for uri in uris or [] for f in self._filters(uri) if f]
        with self._connect() as c:
            tracks = schema.search_tracks(c, q, limit, offset, exact, filters)
        uri = uritools.uricompose("local", path="search", query=q)
        return SearchResult(uri=uri, tracks=tracks)

    def get_images(self, uris):
        images = {}
        with self._connect() as c:
            for uri in uris:
                if uri.startswith("local:album"):
                    images[uri] = schema.get_album_images(c, uri)
                elif uri.startswith("local:track"):
                    images[uri] = schema.get_track_images(c, uri)
        return images

    def get_distinct(self, field, query=None):
        q = []
        for key, values in query.items() if query else []:
            q.extend((key, value) for value in values)
        # Gracefully handle both old and new field values for this API.
        compat_field = {"track": "track_name"}.get(field, field)
        return set(schema.list_distinct(self._connect(), compat_field, q))

    def _connect(self):
        if not self._connection:
            self._connection = sqlite3.connect(
                self._dbpath,
                factory=schema.Connection,
                timeout=self._config["timeout"],
                check_same_thread=False,
            )
        return self._connection

    def _browse_album(self, uri, order=("disc_no", "track_no", "name")):
        return schema.browse(self._connect(), Ref.TRACK, order, album=uri)

    def _browse_artist(self, uri, order=("type", "name COLLATE NOCASE")):
        with self._connect() as c:
            albums = schema.browse(c, Ref.ALBUM, order, albumartist=uri)
            refs = schema.browse(c, order=order, artist=uri)
        album_uris, tracks = {ref.uri for ref in albums}, []
        for ref in refs:
            if ref.type == Ref.ALBUM and ref.uri not in album_uris:
                albums.append(
                    Ref.directory(
                        uri=uritools.uricompose(
                            "local",
                            None,
                            "directory",
                            dict(type=Ref.TRACK, album=ref.uri, artist=uri),
                        ),
                        name=ref.name,
                    )
                )
            elif ref.type == Ref.TRACK:
                tracks.append(ref)
            else:
                logger.debug("Skipped SQLite browse result %s", ref.uri)
        albums.sort(key=operator.attrgetter("name"))
        return albums + tracks

    def _browse_directory(self, uri, order=("type", "name COLLATE NOCASE")):
        query = dict(uritools.urisplit(uri).getquerylist())
        type = query.pop("type", None)
        role = query.pop("role", None)

        # TODO: handle these in schema (generically)?
        if type == "date":
            format = query.get("format", "%Y-%m-%d")
            return list(map(date_ref, schema.dates(self._connect(), format=format)))
        if type == "genre":
            return list(map(genre_ref, schema.list_distinct(self._connect(), "genre")))

        # Fix #38: keep sort order of album tracks; this also applies
        # to composers and performers
        if type == Ref.TRACK and "album" in query:
            order = ("disc_no", "track_no", "name")
        if type == Ref.ARTIST and self._config["use_artist_sortname"]:
            order = ("coalesce(sortname, name) COLLATE NOCASE",)
        roles = role or ("artist", "albumartist")  # FIXME: re-think 'roles'...

        refs = []
        for ref in schema.browse(
            self._connect(), type, order, role=roles, **query
        ):  # noqa
            if ref.type == Ref.TRACK or (not query and not role):
                refs.append(ref)
            elif ref.type == Ref.ALBUM:
                refs.append(
                    Ref.directory(
                        uri=uritools.uricompose(
                            "local",
                            None,
                            "directory",
                            dict(query, type=Ref.TRACK, album=ref.uri),  # noqa
                        ),
                        name=ref.name,
                    )
                )
            elif ref.type == Ref.ARTIST:
                refs.append(
                    Ref.directory(
                        uri=uritools.uricompose(
                            "local", None, "directory", dict(query, **{role: ref.uri})
                        ),
                        name=ref.name,
                    )
                )
            else:
                logger.warning("Unexpected SQLite browse result: %r", ref)
        return refs

    def _filters(self, uri):
        if uri.startswith("local:directory"):
            return [dict(uritools.urisplit(uri).getquerylist())]
        elif uri.startswith("local:artist"):
            return [{"artist": uri}, {"albumartist": uri}]
        elif uri.startswith("local:album"):
            return [{"album": uri}]
        else:
            return []



import time
import unittest

import pykka
from mopidy import core
from mopidy.core import PlaybackState
from mopidy.models import TlTrack, Track

from mopidy_local import actor
from unittest import mock
from tests import dummy_audio, generate_song, path_to_data_dir, populate_tracklist

# TODO Test 'playlist repeat', e.g. repeat=1,single=0


class LocalPlaybackProviderTest(unittest.TestCase):
    config = {
        "core": {"data_dir": path_to_data_dir(""), "max_tracklist_length": 10000},
        "local": {
            "media_dir": path_to_data_dir(""),
            "directories": [],
            "timeout": 10,
            "use_artist_sortname": False,
            "album_art_files": [],
        },
    }

    # We need four tracks so that our shuffled track tests behave nicely with
    # reversed as a fake shuffle. Ensuring that shuffled order is [4,3,2,1] and
    # normal order [1,2,3,4] which means next_track != next_track_with_random
    tracks = [Track(uri=generate_song(i), length=4464) for i in (1, 2, 3, 4)]

    def add_track(self, uri):
        track = Track(uri=uri, length=4464)
        self.tracklist.add([track])

    def trigger_about_to_finish(self):
        # Flush any queued core calls.
        self.playback.get_current_tl_track().get()

        callback = self.audio.get_about_to_finish_callback().get()
        callback()

    def setUp(self):  # noqa: N802
        self.audio = dummy_audio.create_proxy()
        self.backend = actor.LocalBackend.start(
            config=self.config, audio=self.audio
        ).proxy()
        self.core = core.Core.start(
            audio=self.audio, backends=[self.backend], config=self.config
        ).proxy()
        self.playback = self.core.playback
        self.tracklist = self.core.tracklist

        assert len(self.tracks) >= 3, "Need at least three tracks to run tests."
        assert (
            self.tracks[0].length >= 2000
        ), "First song needs to be at least 2000 miliseconds"

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def assert_state_is(self, state):
        assert self.playback.get_state().get() == state

    def assert_current_track_is(self, track):
        assert self.playback.get_current_track().get() == track

    def assert_current_track_is_not(self, track):
        assert self.playback.get_current_track().get() != track

    def assert_current_track_index_is(self, index):
        assert self.tracklist.index().get() == index

    def assert_next_tl_track_is(self, tl_track):
        next_tlid = self.tracklist.get_next_tlid().get()
        assert next_tlid == (tl_track and tl_track.tlid)

    def assert_next_tl_track_is_not(self, tl_track):
        next_tlid = self.tracklist.get_next_tlid().get()
        assert next_tlid != (tl_track and tl_track.tlid)

    def assert_previous_tl_track_is(self, tl_track):
        previous_tlid = self.tracklist.get_previous_tlid().get()
        assert previous_tlid == (tl_track and tl_track.tlid)

    def assert_eot_tl_track_is(self, tl_track):
        eot_tlid = self.tracklist.get_eot_tlid().get()
        assert eot_tlid == (tl_track and tl_track.tlid)

    def assert_eot_tl_track_is_not(self, tl_track):
        eot_tlid = self.tracklist.get_eot_tlid().get()
        assert eot_tlid != (tl_track and tl_track.tlid)

    def test_uri_scheme(self):
        assert "file" not in self.core.get_uri_schemes().get()
        assert "local" in self.core.get_uri_schemes().get()

    def test_play_mp3(self):
        self.add_track("local:track:blank.mp3")
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_play_ogg(self):
        self.add_track("local:track:blank.ogg")
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_play_flac(self):
        self.add_track("local:track:blank.flac")
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_play_uri_with_non_ascii_bytes(self):
        # Regression test: If trying to do .split(u':') on a bytestring, the
        # string will be decoded from ASCII to Unicode, which will crash on
        # non-ASCII strings, like the bytestring the following URI decodes to.
        self.add_track("local:track:12%20Doin%E2%80%99%20It%20Right.flac")
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    def test_initial_state_is_stopped(self):
        self.assert_state_is(PlaybackState.STOPPED)

    def test_play_with_empty_playlist(self):
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.STOPPED)

    def test_play_with_empty_playlist_return_value(self):
        assert self.playback.play().get() is None

    @populate_tracklist
    def test_play_state(self):
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_play_return_value(self):
        assert self.playback.play().get() is None

    @populate_tracklist
    def test_play_track_state(self):
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_play_track_return_value(self):
        assert self.playback.play(self.tl_tracks.get()[(-1)]).get() is None

    @populate_tracklist
    def test_play_when_playing(self):
        self.playback.play().get()
        track = self.playback.get_current_track().get()
        self.playback.play().get()
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_play_when_paused(self):
        self.playback.play().get()
        track = self.playback.get_current_track().get()
        self.playback.pause().get()
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_play_when_paused_after_next(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.next().get()
        track = self.playback.get_current_track().get()
        self.playback.pause().get()
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_play_sets_current_track(self):
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_play_track_sets_current_track(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.assert_current_track_is(self.tracks[-1])

    @populate_tracklist
    def test_play_skips_to_next_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        uri = self.backend.playback.translate_uri(self.tracks[0].uri).get()
        self.audio.trigger_fake_playback_failure(uri)

        self.playback.play().get()
        self.assert_current_track_is_not(self.tracks[0])
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_current_track_after_completed_playlist(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.playback.next().get()
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_previous(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.previous().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_previous_more(self):
        self.playback.play().get()  # At track 0
        self.playback.next().get()  # At track 1
        self.playback.next().get()  # At track 2
        self.playback.previous().get()  # At track 1
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_previous_return_value(self):
        self.playback.play().get()
        self.playback.next().get()
        assert self.playback.previous().get() is None

    @populate_tracklist
    def test_previous_does_not_trigger_playback(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.stop()
        self.playback.previous().get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_previous_at_start_of_playlist(self):
        self.playback.previous().get()
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    def test_previous_for_empty_playlist(self):
        self.playback.previous().get()
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_previous_skips_to_previous_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        uri = self.backend.playback.translate_uri(self.tracks[1].uri).get()
        self.audio.trigger_fake_playback_failure(uri)

        self.playback.play(self.tl_tracks.get()[2]).get()
        self.assert_current_track_is(self.tracks[2])
        self.playback.previous().get()
        self.assert_current_track_is_not(self.tracks[1])
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_next(self):
        self.playback.play().get()

        old_track = self.playback.get_current_track().get()
        old_position = self.tracklist.index().get()

        self.playback.next().get()

        assert self.tracklist.index().get() == (old_position + 1)
        self.assert_current_track_is_not(old_track)

    @populate_tracklist
    def test_next_return_value(self):
        self.playback.play().get()
        assert self.playback.next().get() is None

    @populate_tracklist
    def test_next_does_not_trigger_playback(self):
        self.playback.next().get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_next_at_end_of_playlist(self):
        self.playback.play().get()

        for i, track in enumerate(self.tracks):
            self.assert_state_is(PlaybackState.PLAYING)
            self.assert_current_track_is(track)
            assert self.tracklist.index().get() == i

            self.playback.next()  # noqa: B305

        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_next_until_end_of_playlist_and_play_from_start(self):
        self.playback.play().get()

        for _ in self.tracks:
            self.playback.next().get()

        self.assert_current_track_is(None)
        self.assert_state_is(PlaybackState.STOPPED)

        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(self.tracks[0])

    def test_next_for_empty_playlist(self):
        self.playback.next().get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_next_skips_to_next_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        uri = self.backend.playback.translate_uri(self.tracks[1].uri).get()
        self.audio.trigger_fake_playback_failure(uri)

        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.playback.next().get()
        self.assert_current_track_is_not(self.tracks[1])
        self.assert_current_track_is(self.tracks[2])

    @populate_tracklist
    def test_next_track_before_play(self):
        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    def test_next_track_during_play(self):
        self.playback.play().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    @populate_tracklist
    def test_next_track_after_previous(self):
        self.playback.play().get()
        self.playback.next().get()
        self.playback.previous().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    def test_next_track_empty_playlist(self):
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_next_track_at_end_of_playlist(self):
        self.playback.play().get()
        for _ in self.tl_tracks.get()[1:]:
            self.playback.next().get()
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_next_track_at_end_of_playlist_with_repeat(self):
        self.tracklist.set_repeat(True)
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.playback.next().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_next_track_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)
        self.assert_next_tl_track_is(self.tl_tracks.get()[-1])

    @populate_tracklist
    def test_next_with_consume(self):
        self.tracklist.set_consume(True)
        self.playback.play().get()
        self.playback.next().get()
        assert self.tracks[0] not in self.tracklist.get_tracks().get()

    @populate_tracklist
    def test_next_with_single_and_repeat(self):
        self.tracklist.set_single(True)
        self.tracklist.set_repeat(True)
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.playback.next().get()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_next_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[-1])
        self.playback.next().get()
        self.assert_current_track_is(self.tracks[-2])

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_next_track_with_random_after_append_playlist(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)

        expected_tl_track = self.tl_tracks.get()[-1]
        next_tlid = self.tracklist.get_next_tlid().get()

        # Baseline checking that first next_track is last tl track per our fake
        # shuffle.
        assert next_tlid == expected_tl_track.tlid

        self.tracklist.add(self.tracks[:1])

        old_next_tlid = next_tlid
        expected_tl_track = self.tracklist.get_tl_tracks().get()[-1]
        next_tlid = self.tracklist.get_next_tlid().get()

        # Verify that first next track has changed since we added to the
        # playlist.
        assert next_tlid == expected_tl_track.tlid
        assert next_tlid != old_next_tlid

    @populate_tracklist
    def test_end_of_track(self):
        self.playback.play().get()

        old_track = self.playback.get_current_track().get()
        old_position = self.tracklist.index().get()

        self.trigger_about_to_finish()

        new_track = self.playback.get_current_track().get()
        assert self.tracklist.index().get() == (old_position + 1)
        assert new_track.uri != old_track.uri

    @populate_tracklist
    def test_end_of_track_return_value(self):
        self.playback.play().get()
        assert self.trigger_about_to_finish() is None

    @populate_tracklist
    def test_end_of_track_does_not_trigger_playback(self):
        self.trigger_about_to_finish()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_track_at_end_of_playlist(self):
        self.playback.play().get()

        for i, track in enumerate(self.tracks):
            self.assert_state_is(PlaybackState.PLAYING)
            self.assert_current_track_is(track)
            assert self.tracklist.index().get() == i

            self.trigger_about_to_finish()

        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_track_until_end_of_playlist_and_play_from_start(self):
        self.playback.play().get()

        for _ in self.tracks:
            self.trigger_about_to_finish()

        assert self.playback.get_current_track().get() is None
        self.assert_state_is(PlaybackState.STOPPED)

        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(self.tracks[0])

    def test_end_of_track_for_empty_playlist(self):
        self.trigger_about_to_finish()
        self.assert_state_is(PlaybackState.STOPPED)

    # TODO: On about to finish does not handle skipping to next track yet.
    @unittest.expectedFailure
    @populate_tracklist
    def test_end_of_track_skips_to_next_track_on_failure(self):
        # If backend's play() returns False, it is a failure.
        return_values = [True, False, True]
        self.backend.playback.play = lambda: return_values.pop()
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.trigger_about_to_finish()
        self.assert_current_track_is_not(self.tracks[1])
        self.assert_current_track_is(self.tracks[2])

    @populate_tracklist
    def test_end_of_track_track_before_play(self):
        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    def test_end_of_track_track_during_play(self):
        self.playback.play().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    @populate_tracklist
    def test_about_to_finish_after_previous(self):
        self.playback.play().get()
        self.trigger_about_to_finish()
        self.playback.previous().get()
        self.assert_next_tl_track_is(self.tl_tracks.get()[1])

    def test_end_of_track_track_empty_playlist(self):
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_end_of_track_track_at_end_of_playlist(self):
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.trigger_about_to_finish()

        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_end_of_track_track_at_end_of_playlist_with_repeat(self):
        self.tracklist.set_repeat(True)
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.trigger_about_to_finish()

        self.assert_next_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_end_of_track_track_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)
        self.assert_next_tl_track_is(self.tl_tracks.get()[-1])

    @populate_tracklist
    def test_end_of_track_with_consume(self):
        self.tracklist.set_consume(True)
        self.playback.play().get()
        self.trigger_about_to_finish()
        assert self.tracks[0] not in self.tracklist.get_tracks().get()

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_end_of_track_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[-1])
        self.trigger_about_to_finish()
        self.assert_current_track_is(self.tracks[-2])

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_end_of_track_track_with_random_after_append_playlist(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)

        expected_tl_track = self.tracklist.get_tl_tracks().get()[-1]
        eot_tlid = self.tracklist.get_eot_tlid().get()

        # Baseline checking that first eot_track is last tl track per our fake
        # shuffle.
        assert eot_tlid == expected_tl_track.tlid

        self.tracklist.add(self.tracks[:1])

        old_eot_tlid = eot_tlid
        expected_tl_track = self.tracklist.get_tl_tracks().get()[-1]
        eot_tlid = self.tracklist.get_eot_tlid().get()

        # Verify that first next track has changed since we added to the
        # playlist.
        assert eot_tlid == expected_tl_track.tlid
        assert eot_tlid != old_eot_tlid

    @populate_tracklist
    def test_previous_track_before_play(self):
        self.assert_previous_tl_track_is(None)

    @populate_tracklist
    def test_previous_track_after_play(self):
        self.playback.play().get()
        self.assert_previous_tl_track_is(None)

    @populate_tracklist
    def test_previous_track_after_next(self):
        self.playback.play().get()
        self.playback.next().get()
        self.assert_previous_tl_track_is(self.tl_tracks.get()[0])

    @populate_tracklist
    def test_previous_track_after_previous(self):
        self.playback.play().get()  # At track 0
        self.playback.next().get()  # At track 1
        self.playback.next().get()  # At track 2
        self.playback.previous().get()  # At track 1
        self.assert_previous_tl_track_is(self.tl_tracks.get()[0])

    def test_previous_track_empty_playlist(self):
        self.assert_previous_tl_track_is(None)

    @populate_tracklist
    def test_previous_track_with_consume(self):
        self.tracklist.set_consume(True)
        for _ in self.tracks:
            self.playback.next()  # noqa: B305
            current = self.playback.get_current_tl_track().get()
            self.assert_previous_tl_track_is(current)

    @populate_tracklist
    def test_previous_track_with_random(self):
        self.tracklist.set_random(True)
        for _ in self.tracks:
            self.playback.next()  # noqa: B305
            current = self.playback.get_current_tl_track().get()
            self.assert_previous_tl_track_is(current)

    @populate_tracklist
    def test_initial_current_track(self):
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_current_track_during_play(self):
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_current_track_after_next(self):
        self.playback.play()
        self.playback.next().get()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_initial_tracklist_position(self):
        assert self.tracklist.index().get() is None

    @populate_tracklist
    def test_tracklist_position_during_play(self):
        self.playback.play().get()
        self.assert_current_track_index_is(0)

    @populate_tracklist
    def test_tracklist_position_after_next(self):
        self.playback.play().get()
        self.playback.next().get()
        self.assert_current_track_index_is(1)

    @populate_tracklist
    def test_tracklist_position_at_end_of_playlist(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_current_track_index_is(None)

    @mock.patch("mopidy.core.playback.PlaybackController._on_tracklist_change")
    def test_on_tracklist_change_gets_called(self, change_mock):
        self.tracklist.add([Track()]).get()
        change_mock.assert_called_once_with()

    @populate_tracklist
    def test_on_tracklist_change_when_playing(self):
        self.playback.play().get()
        current_track = self.playback.get_current_track().get()
        self.tracklist.add([self.tracks[2]])
        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_on_tracklist_change_when_stopped(self):
        self.tracklist.add([self.tracks[2]])
        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_on_tracklist_change_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        current_track = self.playback.get_current_track().get()
        self.tracklist.add([self.tracks[2]])
        self.assert_state_is(PlaybackState.PAUSED)
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_pause_when_stopped(self):
        self.playback.pause()
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_pause_when_playing(self):
        self.playback.play().get()
        self.playback.pause()
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_pause_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        self.playback.pause()
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_pause_return_value(self):
        self.playback.play().get()
        assert self.playback.pause().get() is None

    @populate_tracklist
    def test_resume_when_stopped(self):
        self.playback.resume()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_resume_when_playing(self):
        self.playback.play().get()
        self.playback.resume()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_resume_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        self.playback.resume()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_resume_return_value(self):
        self.playback.play().get()
        self.playback.pause()
        assert self.playback.resume().get() is None

    @unittest.SkipTest  # Uses sleep and might not work with LocalBackend
    @populate_tracklist
    def test_resume_continues_from_right_position(self):
        self.playback.play().get()
        time.sleep(0.2)
        self.playback.pause()
        self.playback.resume()
        assert self.playback.get_time_position() != 0

    @populate_tracklist
    def test_seek_when_stopped(self):
        result = self.playback.seek(1000)
        assert result

    @unittest.SkipTest  # tkem doesn't know what's going on here
    @populate_tracklist
    def test_seek_when_stopped_updates_position(self):
        self.playback.seek(1000).get()
        position = self.playback.get_time_position()
        assert position >= 990

    def test_seek_on_empty_playlist(self):
        assert not self.playback.seek(0).get()

    def test_seek_on_empty_playlist_updates_position(self):
        self.playback.seek(0).get()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_seek_when_stopped_triggers_play(self):
        self.playback.seek(0).get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_seek_when_playing(self):
        self.playback.play().get()
        result = self.playback.seek(self.tracks[0].length - 1000)
        assert result

    @populate_tracklist
    def test_seek_when_playing_updates_position(self):
        length = self.tracks[0].length
        self.playback.play().get()
        self.playback.seek(length - 1000).get()
        position = self.playback.get_time_position().get()
        assert position >= (length - 1010)

    @populate_tracklist
    def test_seek_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        result = self.playback.seek(self.tracks[0].length - 1000)
        assert result
        self.assert_state_is(PlaybackState.PAUSED)

    @populate_tracklist
    def test_seek_when_paused_updates_position(self):
        length = self.tracks[0].length
        self.playback.play().get()
        self.playback.pause()
        self.playback.seek(length - 1000)
        position = self.playback.get_time_position().get()
        assert position >= (length - 1010)

    @unittest.SkipTest
    @populate_tracklist
    def test_seek_beyond_end_of_song(self):
        # FIXME need to decide return value
        self.playback.play().get()
        result = self.playback.seek(self.tracks[0].length * 100)
        assert not result

    @populate_tracklist
    def test_seek_beyond_end_of_song_jumps_to_next_song(self):
        self.playback.play().get()
        self.playback.seek(self.tracks[0].length * 100).get()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_seek_beyond_end_of_song_for_last_track(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.playback.seek(self.tracks[-1].length * 100)
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_stop_when_stopped(self):
        self.playback.stop()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_stop_when_playing(self):
        self.playback.play().get()
        self.playback.stop()
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_stop_when_paused(self):
        self.playback.play().get()
        self.playback.pause()
        self.playback.stop()
        self.assert_state_is(PlaybackState.STOPPED)

    def test_stop_return_value(self):
        self.playback.play().get()
        assert self.playback.stop().get() is None

    def test_time_position_when_stopped(self):
        assert self.playback.get_time_position().get() == 0

    @populate_tracklist
    def test_time_position_when_stopped_with_playlist(self):
        assert self.playback.get_time_position().get() == 0

    @unittest.SkipTest  # Uses sleep and does might not work with LocalBackend
    @populate_tracklist
    def test_time_position_when_playing(self):
        self.playback.play().get()
        first = self.playback.get_time_position().get()
        time.sleep(1)
        second = self.playback.get_time_position().get()
        assert second > first

    @populate_tracklist
    def test_time_position_when_paused(self):
        self.playback.play().get()
        self.playback.pause().get()
        first = self.playback.get_time_position().get()
        second = self.playback.get_time_position().get()
        assert first == second

    @populate_tracklist
    def test_play_with_consume(self):
        self.tracklist.set_consume(True)
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_playlist_is_empty_after_all_tracks_are_played_with_consume(self):
        self.tracklist.set_consume(True)
        self.playback.play().get()

        for _ in self.tracks:
            self.trigger_about_to_finish()
        # EOS should have trigger

        assert len(self.tracklist.get_tracks().get()) == 0

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_play_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[-1])

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_previous_with_random(self, shuffle_mock):
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        self.tracklist.set_random(True)
        self.playback.play().get()
        self.playback.next().get()
        current_track = self.playback.get_current_track().get()
        self.playback.previous()
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_end_of_song_starts_next_track(self):
        self.playback.play().get()
        self.trigger_about_to_finish()
        self.assert_current_track_is(self.tracks[1])

    @populate_tracklist
    def test_end_of_song_with_single_and_repeat_starts_same(self):
        self.tracklist.set_single(True)
        self.tracklist.set_repeat(True)
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.trigger_about_to_finish()
        self.assert_current_track_is(self.tracks[0])

    @populate_tracklist
    def test_end_of_song_with_single_random_and_repeat_starts_same(self):
        self.tracklist.set_single(True)
        self.tracklist.set_repeat(True)
        self.tracklist.set_random(True)
        self.playback.play().get()
        current_track = self.playback.get_current_track().get()
        self.trigger_about_to_finish()
        self.assert_current_track_is(current_track)

    @populate_tracklist
    def test_end_of_song_with_single_stops(self):
        self.tracklist.set_single(True)
        self.playback.play().get()
        self.assert_current_track_is(self.tracks[0])
        self.trigger_about_to_finish()
        self.assert_current_track_is(None)
        # EOS should have triggered
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_song_with_single_and_random_stops(self):
        self.tracklist.set_single(True)
        self.tracklist.set_random(True)
        self.playback.play().get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_current_track_is(None)
        self.assert_state_is(PlaybackState.STOPPED)

    @populate_tracklist
    def test_end_of_playlist_stops(self):
        self.playback.play(self.tl_tracks.get()[-1]).get()
        self.trigger_about_to_finish()
        # EOS should have triggered
        self.assert_state_is(PlaybackState.STOPPED)

    def test_repeat_off_by_default(self):
        assert self.tracklist.get_repeat().get() is False

    def test_random_off_by_default(self):
        assert self.tracklist.get_random().get() is False

    def test_consume_off_by_default(self):
        assert self.tracklist.get_consume().get() is False

    @populate_tracklist
    def test_random_until_end_of_playlist(self):
        self.tracklist.set_random(True)
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.playback.next().get()
        self.assert_next_tl_track_is(None)

    @populate_tracklist
    def test_random_with_eot_until_end_of_playlist(self):
        self.tracklist.set_random(True)
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.trigger_about_to_finish()

        self.assert_eot_tl_track_is(None)

    @populate_tracklist
    def test_random_until_end_of_playlist_and_play_from_start(self):
        self.tracklist.set_random(True)
        self.playback.play().get()
        for _ in self.tracks:
            self.playback.next().get()
        self.assert_next_tl_track_is_not(None)
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_random_with_eot_until_end_of_playlist_and_play_from_start(self):
        self.tracklist.set_random(True)
        self.playback.play().get()
        for _ in self.tracks:
            self.trigger_about_to_finish()
        # EOS should have triggered

        self.assert_eot_tl_track_is_not(None)
        self.assert_state_is(PlaybackState.STOPPED)
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)

    @populate_tracklist
    def test_random_until_end_of_playlist_with_repeat(self):
        self.tracklist.set_repeat(True)
        self.tracklist.set_random(True)
        self.playback.play().get()
        for _ in self.tracks[1:]:
            self.playback.next()  # noqa: B305
        self.assert_next_tl_track_is_not(None)

    @populate_tracklist
    def test_played_track_during_random_not_played_again(self):
        self.tracklist.set_random(True)
        self.playback.play().get()
        played = []
        for _ in self.tracks:
            track = self.playback.get_current_track().get()
            assert track not in played
            played.append(track)
            self.playback.next().get()

    @populate_tracklist
    @mock.patch("random.shuffle")
    def test_play_track_then_enable_random(self, shuffle_mock):
        # Covers underlying issue IssueGH17RegressionTest tests for.
        shuffle_mock.side_effect = lambda tracks: tracks.reverse()

        expected = self.tl_tracks.get()[::-1] + [None]
        actual = []

        self.playback.play().get()
        self.tracklist.set_random(True)
        while self.playback.get_state().get() != PlaybackState.STOPPED:
            self.playback.next().get()
            actual.append(self.playback.get_current_tl_track().get())
            if len(actual) > len(expected):
                break
        assert actual == expected

    @populate_tracklist
    def test_playing_track_that_isnt_in_playlist(self):
        with self.assertRaises(AssertionError):
            self.playback.play(TlTrack(17, Track())).get()



from mopidy import backend

from mopidy_local import translator


class LocalPlaybackProvider(backend.PlaybackProvider):
    def translate_uri(self, uri):
        return translator.local_uri_to_file_uri(
            uri, self.backend.config["local"]["media_dir"]
        )



import random
import unittest

import pykka
from mopidy import core
from mopidy.core import PlaybackState
from mopidy.models import Playlist, Track

from mopidy_local import actor
from tests import dummy_audio, generate_song, path_to_data_dir, populate_tracklist


class LocalTracklistProviderTest(unittest.TestCase):
    config = {
        "core": {"data_dir": path_to_data_dir(""), "max_tracklist_length": 10000},
        "local": {
            "media_dir": path_to_data_dir(""),
            "directories": [],
            "timeout": 10,
            "use_artist_sortname": False,
            "album_art_files": [],
        },
    }
    tracks = [Track(uri=generate_song(i), length=4464) for i in range(1, 4)]

    def setUp(self):  # noqa: N802
        self.audio = dummy_audio.create_proxy()
        self.backend = actor.LocalBackend.start(
            config=self.config, audio=self.audio
        ).proxy()
        self.core = core.Core.start(
            audio=self.audio, backends=[self.backend], config=self.config
        ).proxy()
        self.controller = self.core.tracklist
        self.playback = self.core.playback

        assert len(self.tracks) == 3, "Need three tracks to run tests."

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()

    def assert_state_is(self, state):
        assert self.playback.get_state().get() == state

    def assert_current_track_is(self, track):
        assert self.playback.get_current_track().get() == track

    def test_length(self):
        assert 0 == len(self.controller.get_tl_tracks().get())
        assert 0 == self.controller.get_length().get()
        self.controller.add(self.tracks)
        assert 3 == len(self.controller.get_tl_tracks().get())
        assert 3 == self.controller.get_length().get()

    def test_add(self):
        for track in self.tracks:
            added = self.controller.add([track]).get()
            tracks = self.controller.get_tracks().get()
            tl_tracks = self.controller.get_tl_tracks().get()

            assert track == tracks[(-1)]
            assert added[0] == tl_tracks[(-1)]
            assert track == added[0].track

    def test_add_at_position(self):
        for track in self.tracks[:-1]:
            added = self.controller.add([track], 0).get()
            tracks = self.controller.get_tracks().get()
            tl_tracks = self.controller.get_tl_tracks().get()

            assert track == tracks[0]
            assert added[0] == tl_tracks[0]
            assert track == added[0].track

    @populate_tracklist
    def test_add_at_position_outside_of_playlist(self):
        for track in self.tracks:
            added = self.controller.add([track], len(self.tracks) + 2).get()
            tracks = self.controller.get_tracks().get()
            tl_tracks = self.controller.get_tl_tracks().get()

            assert track == tracks[(-1)]
            assert added[0] == tl_tracks[(-1)]
            assert track == added[0].track

    @populate_tracklist
    def test_filter_by_tlid(self):
        tl_track = self.controller.get_tl_tracks().get()[1]
        result = self.controller.filter({"tlid": [tl_track.tlid]}).get()
        assert [tl_track] == result

    @populate_tracklist
    def test_filter_by_uri(self):
        tl_track = self.controller.get_tl_tracks().get()[1]
        result = self.controller.filter({"uri": [tl_track.track.uri]}).get()
        assert [tl_track] == result

    @populate_tracklist
    def test_filter_by_uri_returns_nothing_for_invalid_uri(self):
        assert [] == self.controller.filter({"uri": ["foobar"]}).get()

    def test_filter_by_uri_returns_single_match(self):
        t = Track(uri="a")
        self.controller.add([Track(uri="z"), t, Track(uri="y")])

        result = self.controller.filter({"uri": ["a"]}).get()
        assert t == result[0].track

    def test_filter_by_uri_returns_multiple_matches(self):
        track = Track(uri="a")
        self.controller.add([Track(uri="z"), track, track])
        tl_tracks = self.controller.filter({"uri": ["a"]}).get()
        assert track == tl_tracks[0].track
        assert track == tl_tracks[1].track

    def test_filter_by_uri_returns_nothing_if_no_match(self):
        self.controller.playlist = Playlist(tracks=[Track(uri="z"), Track(uri="y")])
        assert [] == self.controller.filter({"uri": ["a"]}).get()

    def test_filter_by_multiple_criteria_returns_elements_matching_all(self):
        t1 = Track(uri="a", name="x")
        t2 = Track(uri="b", name="x")
        t3 = Track(uri="b", name="y")
        self.controller.add([t1, t2, t3])

        result1 = self.controller.filter({"uri": ["a"], "name": ["x"]}).get()
        assert t1 == result1[0].track

        result2 = self.controller.filter({"uri": ["b"], "name": ["x"]}).get()
        assert t2 == result2[0].track

        result3 = self.controller.filter({"uri": ["b"], "name": ["y"]}).get()
        assert t3 == result3[0].track

    def test_filter_by_criteria_that_is_not_present_in_all_elements(self):
        track1 = Track()
        track2 = Track(uri="b")
        track3 = Track()

        self.controller.add([track1, track2, track3])
        result = self.controller.filter({"uri": ["b"]}).get()
        assert track2 == result[0].track

    @populate_tracklist
    def test_clear(self):
        self.controller.clear().get()
        assert len(self.controller.get_tracks().get()) == 0

    def test_clear_empty_playlist(self):
        self.controller.clear().get()
        assert len(self.controller.get_tracks().get()) == 0

    @populate_tracklist
    def test_clear_when_playing(self):
        self.playback.play().get()
        self.assert_state_is(PlaybackState.PLAYING)
        self.controller.clear().get()
        self.assert_state_is(PlaybackState.STOPPED)

    def test_add_appends_to_the_tracklist(self):
        self.controller.add([Track(uri="a"), Track(uri="b")])

        tracks = self.controller.get_tracks().get()
        assert len(tracks) == 2

        self.controller.add([Track(uri="c"), Track(uri="d")])

        tracks = self.controller.get_tracks().get()
        assert len(tracks) == 4
        assert tracks[0].uri == "a"
        assert tracks[1].uri == "b"
        assert tracks[2].uri == "c"
        assert tracks[3].uri == "d"

    def test_add_does_not_reset_version(self):
        version = self.controller.get_version().get()
        self.controller.add([])
        assert self.controller.get_version().get() == version

    @populate_tracklist
    def test_add_preserves_playing_state(self):
        self.playback.play().get()

        track = self.playback.get_current_track().get()
        tracks = self.controller.get_tracks().get()
        self.controller.add(tracks[1:2]).get()

        self.assert_state_is(PlaybackState.PLAYING)
        self.assert_current_track_is(track)

    @populate_tracklist
    def test_add_preserves_stopped_state(self):
        tracks = self.controller.get_tracks().get()
        self.controller.add(tracks[1:2]).get()

        self.assert_state_is(PlaybackState.STOPPED)
        self.assert_current_track_is(None)

    @populate_tracklist
    def test_add_returns_the_tl_tracks_that_was_added(self):
        tracks = self.controller.get_tracks().get()

        added = self.controller.add(tracks[1:2]).get()
        tracks = self.controller.get_tracks().get()
        assert added[0].track == tracks[1]

    @populate_tracklist
    def test_move_single(self):
        self.controller.move(0, 0, 2)

        tracks = self.controller.get_tracks().get()
        assert tracks[2] == self.tracks[0]

    @populate_tracklist
    def test_move_group(self):
        self.controller.move(0, 2, 1)

        tracks = self.controller.get_tracks().get()
        assert tracks[1] == self.tracks[0]
        assert tracks[2] == self.tracks[1]

    @populate_tracklist
    def test_moving_track_outside_of_playlist(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.move(0, 0, num_tracks + 5).get()

    @populate_tracklist
    def test_move_group_outside_of_playlist(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.move(0, 2, num_tracks + 5).get()

    @populate_tracklist
    def test_move_group_out_of_range(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.move(num_tracks + 2, num_tracks + 3, 0).get()

    @populate_tracklist
    def test_move_group_invalid_group(self):
        with self.assertRaises(AssertionError):
            self.controller.move(2, 1, 0).get()

    def test_tracks_attribute_is_immutable(self):
        tracks1 = self.controller.get_tracks().get()
        tracks2 = self.controller.get_tracks().get()
        assert id(tracks1) != id(tracks2)

    @populate_tracklist
    def test_remove(self):
        track1 = self.controller.get_tracks().get()[1]
        track2 = self.controller.get_tracks().get()[2]
        version = self.controller.get_version().get()
        self.controller.remove({"uri": [track1.uri]})
        assert version < self.controller.get_version().get()
        assert track1 not in self.controller.get_tracks().get()
        assert track2 == self.controller.get_tracks().get()[1]

    @populate_tracklist
    def test_removing_track_that_does_not_exist_does_nothing(self):
        self.controller.remove({"uri": ["/nonexistant"]}).get()

    def test_removing_from_empty_playlist_does_nothing(self):
        self.controller.remove({"uri": ["/nonexistant"]}).get()

    @populate_tracklist
    def test_remove_lists(self):
        version = self.controller.get_version().get()
        tracks = self.controller.get_tracks().get()
        track0 = tracks[0]
        track1 = tracks[1]
        track2 = tracks[2]

        self.controller.remove({"uri": [track0.uri, track2.uri]})

        tracks = self.controller.get_tracks().get()
        assert version < self.controller.get_version().get()
        assert track0 not in tracks
        assert track2 not in tracks
        assert track1 == tracks[0]

    @populate_tracklist
    def test_shuffle(self):
        random.seed(1)
        self.controller.shuffle()

        shuffled_tracks = self.controller.get_tracks().get()

        assert self.tracks != shuffled_tracks
        assert set(self.tracks) == set(shuffled_tracks)

    @populate_tracklist
    def test_shuffle_subset(self):
        random.seed(1)
        self.controller.shuffle(1, 3)

        shuffled_tracks = self.controller.get_tracks().get()

        assert self.tracks != shuffled_tracks
        assert self.tracks[0] == shuffled_tracks[0]
        assert set(self.tracks) == set(shuffled_tracks)

    @populate_tracklist
    def test_shuffle_invalid_subset(self):
        with self.assertRaises(AssertionError):
            self.controller.shuffle(3, 1).get()

    @populate_tracklist
    def test_shuffle_superset(self):
        num_tracks = len(self.controller.get_tracks().get())
        with self.assertRaises(AssertionError):
            self.controller.shuffle(1, num_tracks + 5).get()

    @populate_tracklist
    def test_shuffle_open_subset(self):
        random.seed(1)
        self.controller.shuffle(1)

        shuffled_tracks = self.controller.get_tracks().get()

        assert self.tracks != shuffled_tracks
        assert self.tracks[0] == shuffled_tracks[0]
        assert set(self.tracks) == set(shuffled_tracks)

    @populate_tracklist
    def test_slice_returns_a_subset_of_tracks(self):
        track_slice = self.controller.slice(1, 3).get()
        assert 2 == len(track_slice)
        assert self.tracks[1] == track_slice[0].track
        assert self.tracks[2] == track_slice[1].track

    @populate_tracklist
    def test_slice_returns_empty_list_if_indexes_outside_tracks_list(self):
        assert 0 == len(self.controller.slice(7, 8).get())
        assert 0 == len(self.controller.slice((-1), 1).get())

    def test_version_does_not_change_when_adding_nothing(self):
        version = self.controller.get_version().get()
        self.controller.add([])
        assert version == self.controller.get_version().get()

    def test_version_increases_when_adding_something(self):
        version = self.controller.get_version().get()
        self.controller.add([Track()])
        assert version < self.controller.get_version().get()



import pathlib
import os
import unittest
from unittest import mock

import pykka
from mopidy import core
from mopidy.models import SearchResult, Track

from mopidy_local import actor, storage, translator
from tests import dummy_audio, path_to_data_dir


class LocalLibraryProviderTest(unittest.TestCase):
    config = {
        "core": {"data_dir": path_to_data_dir(""), "max_tracklist_length": 10000},
        "local": {
            "media_dir": path_to_data_dir(""),
            "directories": [],
            "timeout": 10,
            "max_search_results": 100,
            "use_artist_sortname": False,
            "album_art_files": [],
        },
    }

    def setUp(self):
        self.audio = dummy_audio.create_proxy()
        self.backend = actor.LocalBackend.start(
            config=self.config, audio=self.audio
        ).proxy()
        self.core = core.Core.start(
            audio=self.audio, backends=[self.backend], config=self.config
        ).proxy()
        self.library = self.backend.library
        self.storage = storage.LocalStorageProvider(self.config)
        self.storage.load()

    def tearDown(self):  # noqa: N802
        pykka.ActorRegistry.stop_all()
        try:
            os.remove(path_to_data_dir("local/library.db"))
        except OSError:
            pass

    def test_add_noname_ascii(self):
        name = "Test.mp3"
        uri = translator.path_to_local_track_uri(name, pathlib.Path("/media/dir"))
        track = Track(name=name, uri=uri)
        self.storage.begin()
        self.storage.add(track)
        self.storage.close()
        assert [track] == self.library.lookup(uri).get()

    def test_add_noname_utf8(self):
        name = "Mi\xf0vikudags.mp3"
        uri = translator.path_to_local_track_uri(
            name.encode(), pathlib.Path("/media/dir")
        )
        track = Track(name=name, uri=uri)
        self.storage.begin()
        self.storage.add(track)
        self.storage.close()
        assert [track] == self.library.lookup(uri).get()

    def test_clear(self):
        self.storage.begin()
        self.storage.add(Track(uri="local:track:track.mp3"))
        self.storage.close()
        self.storage.clear()
        assert self.storage.load() == 0

    def test_search_uri(self):
        lib = self.library
        empty = SearchResult(uri="local:search?")
        assert empty == lib.search(uris=None).get()
        assert empty == lib.search(uris=[]).get()
        assert empty == lib.search(uris=["local:"]).get()
        assert empty == lib.search(uris=["local:directory"]).get()
        assert empty == lib.search(uris=["local:directory:"]).get()
        assert empty == lib.search(uris=["foobar:"]).get()

    @mock.patch("mopidy_local.schema.list_distinct")
    def test_distinct_field_track_uses_track_name(self, distinct_mock):
        distinct_mock.return_value = []

        assert self.library.get_distinct("track").get() == set()
        distinct_mock.assert_called_once_with(mock.ANY, "track_name", [])



from __future__ import annotations

import logging
import os
import urllib
from pathlib import Path
from typing import Union

logger = logging.getLogger(__name__)


def local_uri_to_file_uri(local_uri: str, media_dir: Path) -> str:
    """Convert local track or directory URI to file URI."""
    path = local_uri_to_path(local_uri, media_dir)
    return path.as_uri()


def local_uri_to_path(local_uri: str, media_dir: Path) -> Path:
    """Convert local track or directory URI to absolute path."""
    if not local_uri.startswith(("local:directory:", "local:track:")):
        raise ValueError("Invalid URI.")
    uri_path = urllib.parse.urlsplit(local_uri.split(":", 2)[2]).path
    file_bytes = urllib.parse.unquote_to_bytes(uri_path)
    file_path = Path(os.fsdecode(file_bytes))
    return media_dir / file_path


def path_to_file_uri(path: Union[str, bytes, Path]) -> str:
    """Convert absolute path to file URI."""
    ppath = Path(os.fsdecode(path))
    assert ppath.is_absolute()
    return ppath.as_uri()


def path_to_local_track_uri(path: Union[str, bytes, Path], media_dir: Path) -> str:
    """Convert path to local track URI."""
    ppath = Path(os.fsdecode(path))
    if ppath.is_absolute():
        ppath = ppath.relative_to(media_dir)
    quoted_path = urllib.parse.quote(bytes(ppath))
    return f"local:track:{quoted_path}"



import logging
import operator
import pathlib
import re
import sqlite3

from mopidy.models import Album, Artist, Image, Ref, Track

_IMAGE_SIZE_RE = re.compile(r".*-(\d+)x(\d+)\.(?:png|gif|jpeg)$")

_IMAGES_QUERY = "SELECT images FROM album WHERE images IS NOT NULL"

_ALBUM_IMAGE_QUERY = "SELECT images FROM album WHERE uri = ?"

_TRACK_IMAGE_QUERY = """
SELECT album.images AS images
  FROM track
  LEFT OUTER JOIN album ON track.album = album.uri
 WHERE track.uri = ?
"""

_BROWSE_QUERIES = {
    None: """
    SELECT CASE WHEN album.uri IS NULL THEN '%s' ELSE '%s' END AS type,
           coalesce(album.uri, track.uri) AS uri,
           coalesce(album.name, track.name) AS name
      FROM track LEFT OUTER JOIN album ON track.album = album.uri
     WHERE %%s
     GROUP BY coalesce(album.uri, track.uri)
     ORDER BY %%s
    """
    % (Ref.TRACK, Ref.ALBUM),
    Ref.ALBUM: """
    SELECT '%s' AS type, uri AS uri, name AS name
      FROM album
     WHERE %%s
     ORDER BY %%s
    """
    % Ref.ALBUM,
    Ref.ARTIST: """
    SELECT '%s' AS type, uri AS uri, name AS name
      FROM artist
     WHERE %%s
     ORDER BY %%s
    """
    % Ref.ARTIST,
    Ref.TRACK: """
    SELECT '%s' AS type, uri AS uri, name AS name
      FROM track
     WHERE %%s
     ORDER BY %%s
    """
    % Ref.TRACK,
}

_BROWSE_FILTERS = {
    None: {
        "album": "track.album = ?",
        "albumartist": "album.artists = ?",
        "artist": "track.artists = ?",
        "composer": "track.composers = ?",
        "date": "track.date LIKE ? || '%'",
        "genre": "track.genre = ?",
        "performer": "track.performers = ?",
        "max-age": "track.last_modified >= (strftime('%s', 'now') - ?) * 1000",
    },
    Ref.ARTIST: {
        "role": {
            "albumartist": """EXISTS (
                SELECT * FROM album WHERE album.artists = artist.uri
            )""",
            "artist": """EXISTS (
                SELECT * FROM track WHERE track.artists = artist.uri
            )""",
            "composer": """EXISTS (
                SELECT * FROM track WHERE track.composers = artist.uri
            )""",
            "performer": """EXISTS (
                SELECT * FROM track WHERE track.performers = artist.uri
            )""",
        },
    },
    Ref.ALBUM: {
        "albumartist": "artists = ?",
        "artist": """? IN (
            SELECT artists FROM track WHERE album = album.uri
        )""",
        "composer": """? IN (
            SELECT composers FROM track WHERE album = album.uri
        )""",
        "date": """EXISTS (
            SELECT * FROM track WHERE album = album.uri AND date LIKE ? || '%'
        )""",
        "genre": """? IN (
            SELECT genre FROM track WHERE album = album.uri
        )""",
        "performer": """? IN (
            SELECT performers FROM track WHERE album = album.uri
        )""",
        "max-age": """EXISTS (
            SELECT *
              FROM track
             WHERE album = album.uri
               AND last_modified >= (strftime('%s', 'now') - ?) * 1000
        )""",
    },
    Ref.TRACK: {
        "album": "album = ?",
        "albumartist": """? IN (
            SELECT artists FROM album WHERE uri = track.album
        )""",
        "artist": "artists = ?",
        "composer": "composers = ?",
        "date": "date LIKE ? || '%'",
        "genre": "genre = ?",
        "performer": "performers = ?",
        "max-age": "last_modified >= (strftime('%s', 'now') - ?) * 1000",
    },
}

_LOOKUP_QUERIES = {
    Ref.ALBUM: """
    SELECT * FROM tracks WHERE album_uri = ?
    """,
    Ref.ARTIST: """
    SELECT * FROM tracks WHERE ? IN (artist_uri, albumartist_uri)
    """,
    Ref.TRACK: """
    SELECT * FROM tracks WHERE uri = ?
    """,
}

_SEARCH_SQL = """
SELECT *
  FROM tracks
 WHERE docid IN (SELECT docid FROM %s WHERE %s)
"""

_SEARCH_FILTERS = {
    "album": "album_uri = ?",
    "albumartist": "albumartist_uri = ?",
    "artist": "artist_uri = ?",
    "composer": "composer_uri = ?",
    "date": "date LIKE ? || '%'",
    "genre": "genre = ?",
    "performer": "performer_uri = ?",
    "max-age": "last_modified >= (strftime('%s', 'now') - ?) * 1000",
}

_SEARCH_FIELDS = {
    "uri",
    "track_name",
    "album",
    "artist",
    "composer",
    "performer",
    "albumartist",
    "genre",
    "track_no",
    "disc_no",
    "date",
    "comment",
    "musicbrainz_trackid",
    "musicbrainz_albumid",
    "musicbrainz_artistid",
}

schema_version = 7

logger = logging.getLogger(__name__)


class Connection(sqlite3.Connection):
    class Row(sqlite3.Row):
        def __getattr__(self, name):
            return self[name]

    def __init__(self, *args, **kwargs):
        sqlite3.Connection.__init__(self, *args, **kwargs)
        self.execute("PRAGMA foreign_keys = ON")
        self.row_factory = self.Row


def load(c):
    sql_dir = pathlib.Path(__file__).parent / "sql"
    user_version = c.execute("PRAGMA user_version").fetchone()[0]
    while user_version != schema_version:
        if user_version:
            logger.info("Upgrading SQLite database schema v%s", user_version)
            filename = "upgrade-v%s.sql" % user_version
        else:
            logger.info("Creating SQLite database schema v%s", schema_version)
            filename = "schema.sql"
        with open(sql_dir / filename) as fh:
            c.executescript(fh.read())
        new_version = c.execute("PRAGMA user_version").fetchone()[0]
        assert new_version != user_version
        user_version = new_version
    return user_version


def tracks(c):
    return list(map(_track, c.execute("SELECT * FROM tracks")))


def list_distinct(c, field, query=tuple()):
    if field not in _SEARCH_FIELDS:
        raise LookupError("Invalid search field: %s" % field)
    sql = (
        """
    SELECT DISTINCT %s AS field
      FROM search
     WHERE field IS NOT NULL
    """
        % field
    )
    terms = []
    params = []
    for key, value in query:
        if key == "any":
            terms.append("? IN (%s)" % ",".join(_SEARCH_FIELDS))
        elif key in _SEARCH_FIELDS:
            terms.append("%s = ?" % key)
        else:
            raise LookupError("Invalid query field: %s" % key)
        params.append(value)
    if terms:
        sql += " AND " + " AND ".join(terms)
    logger.debug("SQLite list query %r: %s", params, sql)
    return list(map(operator.itemgetter(0), c.execute(sql, params)))


def dates(c, format="%Y-%m-%d"):
    return list(
        map(
            operator.itemgetter(0),
            c.execute(
                """
        SELECT DISTINCT strftime(?, date) AS date
          FROM track
         WHERE date IS NOT NULL
         ORDER BY date
        """,
                [format],
            ),
        )
    )


def lookup(c, type, uri):
    return list(map(_track, c.execute(_LOOKUP_QUERIES[type], [uri])))


def exists(c, uri):
    rows = c.execute("SELECT EXISTS(SELECT * FROM track WHERE uri = ?)", [uri])
    return rows.fetchone()[0]


def browse(c, type=None, order=("type", "name COLLATE NOCASE"), **kwargs):
    filters, params = _filters(_BROWSE_FILTERS[type], **kwargs)
    sql = _BROWSE_QUERIES[type] % (" AND ".join(filters) or "1", ", ".join(order))
    logger.debug("SQLite browse query %r: %s", params, sql)
    return [Ref(**row) for row in c.execute(sql, params)]


def search_tracks(c, query, limit, offset, exact, filters=tuple()):
    if not query:
        sql, params = ("SELECT * FROM tracks WHERE 1", [])
    elif exact:
        sql, params = _indexed_query(query)
    else:
        sql, params = _fulltext_query(query)
    clauses = []
    for kwargs in filters:
        f, p = _filters(_SEARCH_FILTERS, **kwargs)
        if f:
            clauses.append("(%s)" % " AND ".join(f))
            params.extend(p)
        else:
            logger.debug("Skipped SQLite search filter %r", kwargs)
    if clauses:
        sql += " AND (%s)" % " OR ".join(clauses)
    sql += " LIMIT ? OFFSET ?"
    params += [limit, offset]
    logger.debug("SQLite search query %r: %s", params, sql)
    rows = c.execute(sql, params)
    return list(map(_track, rows))


def get_image_uris(c):
    rows = c.execute(_IMAGES_QUERY)
    return (uri for row in rows for uri in row.images.split())


def get_album_images(c, uri):
    images = []
    for row in c.execute(_ALBUM_IMAGE_QUERY, (uri,)):
        images.extend(_images(row.images))
    return images


def get_track_images(c, uri):
    images = []
    for row in c.execute(_TRACK_IMAGE_QUERY, (uri,)):
        images.extend(_images(row.images))
    return images


def insert_artists(c, artists):
    if not artists:
        return None
    if len(artists) != 1:
        logger.warning("Ignoring multiple artists: %r", artists)
    artist = next(iter(artists))
    _insert(
        c,
        "artist",
        {
            "uri": artist.uri,
            "name": artist.name,
            "sortname": artist.sortname,
            "musicbrainz_id": artist.musicbrainz_id,
        },
    )
    return artist.uri


def insert_album(c, album, images=None):
    if not album or not album.name:
        return None
    _insert(
        c,
        "album",
        {
            "uri": album.uri,
            "name": album.name,
            "artists": insert_artists(c, album.artists),
            "num_tracks": album.num_tracks,
            "num_discs": album.num_discs,
            "date": album.date,
            "musicbrainz_id": album.musicbrainz_id,
            "images": " ".join(images) if images else None,
        },
    )
    return album.uri


def insert_track(c, track, images=None):
    _insert(
        c,
        "track",
        {
            "uri": track.uri,
            "name": track.name,
            "album": insert_album(c, track.album, images),
            "artists": insert_artists(c, track.artists),
            "composers": insert_artists(c, track.composers),
            "performers": insert_artists(c, track.performers),
            "genre": track.genre,
            "track_no": track.track_no,
            "disc_no": track.disc_no,
            "date": track.date,
            "length": track.length,
            "bitrate": track.bitrate,
            "comment": track.comment,
            "musicbrainz_id": track.musicbrainz_id,
            "last_modified": track.last_modified,
        },
    )
    return track.uri


def delete_track(c, uri):
    c.execute("DELETE FROM track WHERE uri = ?", (uri,))


def count_tracks(c):
    return c.execute("SELECT count(*) FROM track").fetchone()[0]


def cleanup(c):
    c.execute(
        """
    DELETE FROM album WHERE NOT EXISTS (
        SELECT uri FROM track WHERE track.album = album.uri
    )
    """
    )
    c.execute(
        """
    DELETE FROM artist WHERE NOT EXISTS (
        SELECT uri FROM track WHERE track.artists = artist.uri
         UNION
        SELECT uri FROM track WHERE track.composers = artist.uri
         UNION
        SELECT uri FROM track WHERE track.performers = artist.uri
         UNION
        SELECT uri FROM album WHERE album.artists = artist.uri
    )
    """
    )
    c.execute("ANALYZE")


def clear(c):
    c.executescript(
        """
    DELETE FROM track;
    DELETE FROM album;
    DELETE FROM artist;
    VACUUM;
    """
    )


def _insert(c, table, params):
    sql = "INSERT OR REPLACE INTO {} ({}) VALUES ({})".format(
        table, ", ".join(params.keys()), ", ".join(["?"] * len(params))
    )
    logger.debug("SQLite insert statement: %s %r", sql, params.values())
    return c.execute(sql, list(params.values()))


def _filters(mapping, role=None, **kwargs):
    filters, params = [], []
    if role and "role" in mapping:
        rolemap = mapping["role"]
        if isinstance(role, (str, bytes)):
            filters.append(rolemap[role])
        else:
            filters.append(" OR ".join(rolemap[r] for r in role))
    for key, value in kwargs.items():
        if key in mapping:
            filters.append(mapping[key])
            params.append(value)
        else:
            logger.debug("Skipped SQLite filter expression: %s=%r", key, value)
    return (filters, params)


def _indexed_query(query):
    terms = []
    params = []
    for field, value in query:
        if field == "any":
            terms.append("? IN (%s)" % ",".join(_SEARCH_FIELDS))
        elif field in _SEARCH_FIELDS:
            terms.append("%s = ?" % field)
        else:
            raise LookupError("Invalid search field: %s" % field)
        params.append(value)
    return (_SEARCH_SQL % ("search", " AND ".join(terms)), params)


def _fulltext_query(query):
    terms = []
    params = []
    for field, value in query:
        if field == "any":
            terms.append(_SEARCH_SQL % ("fts", "fts MATCH ?"))
        elif field in _SEARCH_FIELDS:
            terms.append(_SEARCH_SQL % ("fts", "%s MATCH ?" % field))
        else:
            raise LookupError("Invalid search field: %s" % field)
        params.append(value)
    return (" INTERSECT ".join(terms), params)


def _track(row):
    kwargs = {
        "uri": row.uri,
        "name": row.name,
        "genre": row.genre,
        "track_no": row.track_no,
        "disc_no": row.disc_no,
        "date": row.date,
        "length": row.length,
        "bitrate": row.bitrate,
        "comment": row.comment,
        "musicbrainz_id": row.musicbrainz_id,
        "last_modified": row.last_modified,
    }
    if row.album_uri is not None:
        if row.albumartist_uri is not None:
            albumartists = [
                Artist(
                    uri=row.albumartist_uri,
                    name=row.albumartist_name,
                    sortname=row.albumartist_sortname,
                    musicbrainz_id=row.albumartist_musicbrainz_id,
                )
            ]
        else:
            albumartists = None
        kwargs["album"] = Album(
            uri=row.album_uri,
            name=row.album_name,
            artists=albumartists,
            num_tracks=row.album_num_tracks,
            num_discs=row.album_num_discs,
            date=row.album_date,
            musicbrainz_id=row.album_musicbrainz_id,
        )
    if row.artist_uri is not None:
        kwargs["artists"] = [
            Artist(
                uri=row.artist_uri,
                name=row.artist_name,
                sortname=row.artist_sortname,
                musicbrainz_id=row.artist_musicbrainz_id,
            )
        ]
    if row.composer_uri is not None:
        kwargs["composers"] = [
            Artist(
                uri=row.composer_uri,
                name=row.composer_name,
                sortname=row.composer_sortname,
                musicbrainz_id=row.composer_musicbrainz_id,
            )
        ]
    if row.performer_uri is not None:
        kwargs["performers"] = [
            Artist(
                uri=row.performer_uri,
                name=row.performer_name,
                sortname=row.performer_sortname,
                musicbrainz_id=row.performer_musicbrainz_id,
            )
        ]
    return Track(**kwargs)


def _images(field):
    images = []
    for uri in field.split() if field else []:
        m = _IMAGE_SIZE_RE.match(uri)
        if m:
            width = int(m.group(1))
            height = int(m.group(2))
            images.append(Image(uri=uri, width=width, height=height))
        else:
            images.append(Image(uri=uri))
    return images



from mopidy_local import Extension


def test_get_default_config():
    ext = Extension()

    config = ext.get_default_config()

    assert "[local]" in config
    assert "enabled = true" in config


def test_get_config_schema():
    ext = Extension()

    schema = ext.get_config_schema()

    assert "library" in schema
    assert "media_dir" in schema
    assert "data_dir" in schema
    assert "playlists_dir" in schema
    assert "tag_cache_file" in schema
    assert "scan_timeout" in schema
    assert "scan_flush_threshold" in schema
    assert "scan_follow_symlinks" in schema
    assert "included_file_extensions" in schema
    assert "excluded_file_extensions" in schema
    # from mopidy-local-sqlite
    assert "directories" in schema
    assert "timeout" in schema
    assert "use_artist_sortname" in schema
    # from mopidy-local-images
    assert "album_art_files" in schema



import os
import pathlib

import pytest
from mopidy_local import translator


@pytest.mark.parametrize(
    "local_uri,file_uri",
    [
        ("local:directory:A/B", "file:///home/alice/Music/A/B"),
        ("local:directory:A%20B", "file:///home/alice/Music/A%20B"),
        ("local:directory:A+B", "file:///home/alice/Music/A%2BB"),
        (
            "local:directory:%C3%A6%C3%B8%C3%A5",
            "file:///home/alice/Music/%C3%A6%C3%B8%C3%A5",
        ),
        ("local:track:A/B.mp3", "file:///home/alice/Music/A/B.mp3"),
        ("local:track:A%20B.mp3", "file:///home/alice/Music/A%20B.mp3"),
        ("local:track:A+B.mp3", "file:///home/alice/Music/A%2BB.mp3"),
        (
            "local:track:%C3%A6%C3%B8%C3%A5.mp3",
            "file:///home/alice/Music/%C3%A6%C3%B8%C3%A5.mp3",
        ),
    ],
)
def test_local_uri_to_file_uri(local_uri, file_uri):
    media_dir = pathlib.Path("/home/alice/Music")

    assert translator.local_uri_to_file_uri(local_uri, media_dir) == file_uri


@pytest.mark.parametrize("uri", ["A/B", "local:foo:A/B"])
def test_local_uri_to_file_uri_errors(uri):
    media_dir = pathlib.Path("/home/alice/Music")

    with pytest.raises(ValueError):
        translator.local_uri_to_file_uri(uri, media_dir)


@pytest.mark.parametrize(
    "uri,path",
    [
        ("local:directory:A/B", b"/home/alice/Music/A/B"),
        ("local:directory:A%20B", b"/home/alice/Music/A B"),
        ("local:directory:A+B", b"/home/alice/Music/A+B"),
        (
            "local:directory:%C3%A6%C3%B8%C3%A5",
            b"/home/alice/Music/\xc3\xa6\xc3\xb8\xc3\xa5",
        ),
        ("local:track:A/B.mp3", b"/home/alice/Music/A/B.mp3"),
        ("local:track:A%20B.mp3", b"/home/alice/Music/A B.mp3"),
        ("local:track:A+B.mp3", b"/home/alice/Music/A+B.mp3"),
        (
            "local:track:%C3%A6%C3%B8%C3%A5.mp3",
            b"/home/alice/Music/\xc3\xa6\xc3\xb8\xc3\xa5.mp3",
        ),
    ],
)
def test_local_uri_to_path(uri, path):
    media_dir = pathlib.Path("/home/alice/Music")

    result = translator.local_uri_to_path(uri, media_dir)

    assert isinstance(result, pathlib.Path)
    assert bytes(result) == path


@pytest.mark.parametrize("uri", ["A/B", "local:foo:A/B"])
def test_local_uri_to_path_errors(uri):
    media_dir = pathlib.Path("/home/alice/Music")

    with pytest.raises(ValueError):
        translator.local_uri_to_path(uri, media_dir)


@pytest.mark.parametrize(
    "path,uri",
    [
        ("/foo", "file:///foo"),
        (b"/foo", "file:///foo"),
        ("/æøå", "file:///%C3%A6%C3%B8%C3%A5"),
        (b"/\x00\x01\x02", "file:///%00%01%02"),
        (pathlib.Path("/æøå"), "file:///%C3%A6%C3%B8%C3%A5"),
    ],
)
def test_path_to_file_uri(path, uri):
    assert translator.path_to_file_uri(path) == uri


@pytest.mark.parametrize(
    "path,uri",
    [
        (pathlib.Path("foo"), "local:track:foo"),
        (pathlib.Path("/home/alice/Music/foo"), "local:track:foo"),
        (pathlib.Path("æøå"), "local:track:%C3%A6%C3%B8%C3%A5"),
        (pathlib.Path(os.fsdecode(b"\x00\x01\x02")), "local:track:%00%01%02"),
    ],
)
def test_path_to_local_track_uri(path, uri):
    media_dir = pathlib.Path("/home/alice/Music")

    result = translator.path_to_local_track_uri(path, media_dir)

    assert isinstance(result, str)
    assert result == uri



import sqlite3
import unittest

from mopidy.models import Album, Artist, Ref, Track

from mopidy_local import schema

DBPATH = ":memory:"


class SchemaTest(unittest.TestCase):
    artists = [
        Artist(uri="local:artist:0", name="artist #0", musicbrainz_id="1234a-987c"),
        Artist(uri="local:artist:1", name="artist #1"),
    ]
    albums = [
        Album(uri="local:album:0", name="album #0", musicbrainz_id="1234a-3421d"),
        Album(uri="local:album:1", name="album #1", artists=[artists[0]]),
        Album(uri="local:album:2", name="album #2", artists=[artists[1]]),
    ]
    tracks = [
        Track(uri="local:track:0", name="track #0", date="2015-03-15", genre="Rock"),
        Track(uri="local:track:1", name="track #1", artists=[artists[0]]),
        Track(uri="local:track:2", name="track #2", album=albums[0]),
        Track(uri="local:track:3", name="track #3", album=albums[1]),
        Track(
            uri="local:track:4",
            name="track #4",
            album=albums[2],
            composers=[artists[0]],
            performers=[artists[0]],
            musicbrainz_id="1234a-567b",
        ),
    ]

    def setUp(self):
        self.connection = sqlite3.connect(DBPATH, factory=schema.Connection)
        schema.load(self.connection)
        for track in self.tracks:
            schema.insert_track(self.connection, track)

    def tearDown(self):
        self.connection.close()
        self.connection = None

    def test_create(self):
        count = schema.count_tracks(self.connection)
        assert len(self.tracks) == count
        tracks = list(schema.tracks(self.connection))
        assert len(self.tracks) == len(tracks)

    def test_list_distinct(self):
        self.assertEqual(
            [track.name for track in self.tracks],
            schema.list_distinct(self.connection, "track_name"),
        )
        self.assertEqual([], schema.list_distinct(self.connection, "track_no"))
        self.assertEqual([], schema.list_distinct(self.connection, "disc_no"))
        self.assertEqual(
            [album.name for album in self.albums],
            schema.list_distinct(self.connection, "album"),
        )
        self.assertEqual(
            [artist.name for artist in self.artists[0:2]],
            schema.list_distinct(self.connection, "albumartist"),
        )
        self.assertEqual(
            [artist.name for artist in self.artists[0:1]],
            schema.list_distinct(self.connection, "artist"),
        )
        self.assertEqual(
            [artist.name for artist in self.artists[0:1]],
            schema.list_distinct(self.connection, "composer"),
        )
        self.assertEqual(
            [artist.name for artist in self.artists[0:1]],
            schema.list_distinct(self.connection, "performer"),
        )
        self.assertEqual(
            [self.tracks[0].date], schema.list_distinct(self.connection, "date")
        )
        self.assertEqual(
            [self.tracks[0].genre], schema.list_distinct(self.connection, "genre")
        )
        self.assertEqual(
            [self.tracks[4].musicbrainz_id],
            schema.list_distinct(self.connection, "musicbrainz_trackid"),
        )
        self.assertEqual(
            [self.artists[0].musicbrainz_id],
            schema.list_distinct(self.connection, "musicbrainz_artistid"),
        )
        self.assertEqual(
            [self.albums[0].musicbrainz_id],
            schema.list_distinct(self.connection, "musicbrainz_albumid"),
        )

    def test_lookup_track(self):
        with self.connection as c:
            for track in self.tracks:
                result = schema.lookup(c, Ref.TRACK, track.uri)
                assert [track] == list(result)

    def test_lookup_album(self):
        with self.connection as c:
            result = schema.lookup(c, Ref.ALBUM, self.albums[0].uri)
            assert [self.tracks[2]] == list(result)

            result = schema.lookup(c, Ref.ALBUM, self.albums[1].uri)
            assert [self.tracks[3]] == list(result)

            result = schema.lookup(c, Ref.ALBUM, self.albums[2].uri)
            assert [self.tracks[4]] == list(result)

    def test_lookup_artist(self):
        with self.connection as c:
            result = schema.lookup(c, Ref.ARTIST, self.artists[0].uri)
            assert [self.tracks[1], self.tracks[3]] == list(result)

            result = schema.lookup(c, Ref.ARTIST, self.artists[1].uri)
            assert [self.tracks[4]] == list(result)

    @unittest.SkipTest  # TODO: check indexed search
    def test_indexed_search(self):
        for results, query, filters in [
            (map(lambda t: t.uri, self.tracks), [], []),
            ([], [("any", "none")], []),
            (
                [self.tracks[1].uri, self.tracks[3].uri, self.tracks[4].uri],
                [("any", self.artists[0].name)],
                [],
            ),
            (
                [self.tracks[3].uri],
                [("any", self.artists[0].name)],
                [{"album": self.albums[1].uri}],
            ),
            (
                [self.tracks[2].uri],
                [("album", self.tracks[2].album.name)],
                [],
            ),
            (
                [self.tracks[1].uri],
                [("artist", next(iter(self.tracks[1].artists)).name)],
                [],
            ),
            ([self.tracks[0].uri], [("track_name", self.tracks[0].name)], []),
        ]:
            for exact in (True, False):
                with self.connection as c:
                    tracks = schema.search_tracks(c, query, 10, 0, exact, filters)
                self.assertCountEqual(results, map(lambda t: t.uri, tracks))

    def test_fulltext_search(self):
        for results, query, filters in [
            (map(lambda t: t.uri, self.tracks), [("track_name", "track")], []),
            (
                [self.tracks[1].uri, self.tracks[3].uri],
                [("track_name", "track")],
                [
                    {"artist": self.artists[0].uri},
                    {"albumartist": self.artists[0].uri},
                ],
            ),
        ]:
            with self.connection as c:
                tracks = schema.search_tracks(c, query, 10, 0, False, filters)
            self.assertCountEqual(results, map(lambda t: t.uri, tracks))

    def test_browse_artists(self):
        def ref(artist):
            return Ref.artist(name=artist.name, uri=artist.uri)

        with self.connection as c:
            assert list(map(ref, self.artists)) == schema.browse(c, Ref.ARTIST)
            assert list(map(ref, self.artists)) == schema.browse(
                c, Ref.ARTIST, role=["artist", "albumartist"]
            )
            assert list(map(ref, self.artists[0:1])) == schema.browse(
                c, Ref.ARTIST, role="artist"
            )
            assert list(map(ref, self.artists[0:1])) == schema.browse(
                c, Ref.ARTIST, role="composer"
            )
            assert list(map(ref, self.artists[0:1])) == schema.browse(
                c, Ref.ARTIST, role="performer"
            )
            assert list(map(ref, self.artists)) == schema.browse(
                c, Ref.ARTIST, role="albumartist"
            )

    def test_browse_albums(self):
        def ref(album):
            return Ref.album(name=album.name, uri=album.uri)

        with self.connection as c:
            assert list(map(ref, self.albums)) == schema.browse(c, Ref.ALBUM)
            assert list(map(ref, [])) == schema.browse(
                c, Ref.ALBUM, artist=self.artists[0].uri
            )
            assert list(map(ref, self.albums[1:2])) == schema.browse(
                c, Ref.ALBUM, albumartist=self.artists[0].uri
            )

    def test_browse_tracks(self):
        def ref(track):
            return Ref.track(name=track.name, uri=track.uri)

        with self.connection as c:
            assert list(map(ref, self.tracks)) == schema.browse(c, Ref.TRACK)
            assert list(map(ref, self.tracks[1:2])) == schema.browse(
                c, Ref.TRACK, artist=self.artists[0].uri
            )
            assert list(map(ref, self.tracks[2:3])) == schema.browse(
                c, Ref.TRACK, album=self.albums[0].uri
            )
            assert list(map(ref, self.tracks[3:4])) == schema.browse(
                c, Ref.TRACK, albumartist=self.artists[0].uri
            )
            assert list(map(ref, self.tracks[4:5])) == schema.browse(
                c,
                Ref.TRACK,
                composer=self.artists[0].uri,
                performer=self.artists[0].uri,
            )

    def test_delete(self):
        c = self.connection
        schema.delete_track(c, self.tracks[0].uri)
        schema.cleanup(c)
        assert 3 == len(c.execute("SELECT * FROM album").fetchall())
        assert 2 == len(c.execute("SELECT * FROM artist").fetchall())

        schema.delete_track(c, self.tracks[1].uri)
        schema.cleanup(c)
        assert 3 == len(c.execute("SELECT * FROM album").fetchall())
        assert 2 == len(c.execute("SELECT * FROM artist").fetchall())

        schema.delete_track(c, self.tracks[2].uri)
        schema.cleanup(c)
        assert 2 == len(c.execute("SELECT * FROM album").fetchall())
        assert 2 == len(c.execute("SELECT * FROM artist").fetchall())

        schema.delete_track(c, self.tracks[3].uri)
        schema.cleanup(c)
        assert 1 == len(c.execute("SELECT * FROM album").fetchall())
        assert 2 == len(c.execute("SELECT * FROM artist").fetchall())

        schema.delete_track(c, self.tracks[4].uri)
        schema.cleanup(c)
        assert 0 == len(c.execute("SELECT * FROM album").fetchall())
        assert 0 == len(c.execute("SELECT * FROM artist").fetchall())
